from __future__ import annotations
from typing import Union, Type, Tuple
from .device import Device


class ConnectionMetadata:
    """
    Describes the metadata of a connection.
    """

    def __init__(
            self,
            from_device: Union[Type[Device], None] = None,
            to_device: Union[Type[Device], None] = None,
            from_device_node_name: Union[str, None] = None,
            to_device_node_name: Union[str, None] = None,
            bidirectional: bool = True,
    ):

        self._from_device = None
        self._from_device_node_name = None
        self.set_from(from_device, from_device_node_name)

        self._to_device = None
        self._to_device_node_name = None
        self.set_to(to_device, to_device_node_name)

        if not ((from_device is None and to_device is None and from_device_node_name is None
                 and to_device_node_name is None) or (
                from_device is not None and to_device is not None and from_device_node_name is not None and
                to_device_node_name is not None)):
            raise ValueError(
                "you have to provide all or none of the following items: `from_device`, `from_device_node_name`, "
                "`to_device` or `to_device_node_name`")

        # describes if the connection is uni or bidirectional
        self._bidirectional = bidirectional

    def __eq__(self, other: ConnectionMetadata):
        return self.equal_with(other)

    def __hash__(self):
        all_hashes = hash(self._from_device) + hash(self._to_device) + hash(self._from_device_node_name) + \
                     hash(self._to_device_node_name) + hash(self._bidirectional)
        return hash(all_hashes)

    def __compare_with(self, other: ConnectionMetadata, allow_single_unidirectional_for_both_directions: bool) -> bool:
        """
        This method checks, if the metadata of this object is the metadata of the other object.

        The method returns true in the following situations:
        * both connections are bidirectional / the FROM and TO elements (device and node name) are the same
        * both connections are bidirectional / the FROM is the TO and the TO is the FROM
        * both connections are unidirectional and have the same from and to elements

        If the parameter `allow_single_unidirectional_for_both_directions` is True, it additionally checks the following
        situations:
        * one is unidirectional / the other is bidirectional / the FROM and TO elements are the same
        * one is unidirectional / the other is bidirectional / the FROM is the TO and the TO is the FROM
        """
        def check_same() -> bool:
            return (self.from_device == other.from_device and self.from_node_name == other.from_node_name and
                    self.to_device == other.to_device and self.to_node_name == other.to_node_name)

        def check_twisted() -> bool:
            return (self.from_device == other.to_device and self.from_node_name == other.to_node_name and
                    self.to_device == other.from_device and self.to_node_name == other.from_node_name)

        # CHECK: both connections are bidirectional / the FROM and TO elements (device and node name) are the same
        # CHECK: both connections are bidirectional / the FROM is the TO and the TO is the FROM
        if self.bidirectional and other.bidirectional:
            return check_same() or check_twisted()
        # CHECK: both connections are unidirectional and have the same from and to elements
        if not self.bidirectional and not other.bidirectional:
            return check_same()

        if allow_single_unidirectional_for_both_directions:
            # CHECK: one is unidirectional / the other is bidirectional / the FROM and TO elements are the same
            # CHECK: one is unidirectional / the other is bidirectional / the FROM is the TO and the TO is the FROM
            if self.bidirectional and not other.bidirectional or not self.bidirectional and other.bidirectional:
                return check_same() or check_twisted()
        return False

    def set_from(self, from_device: Union[Type[Device], None], from_device_node_name: Union[str, None] = None):
        """
        This method sets the FROM device and node for this connection.

        :param from_device: The FROM device of this connection.
        :param from_device_node_name: The FROM node of this connection (if it should be set, otherwise None).
        """
        if from_device is not None and isinstance(from_device, type) and not issubclass(from_device, Device):
            raise TypeError(f"detect illegal argument element {str(from_device)} for given attribute "
                            f"`from_device` - should be a subclasses of `balder.Device`")
        self._from_device = from_device

        if from_device_node_name is not None and not isinstance(from_device_node_name, str):
            raise TypeError(f"detect illegal argument type {type(from_device_node_name)} for given attribute "
                            f"`from_device_node_name` - should be a string value")
        self._from_device_node_name = from_device_node_name

    def set_to(self, to_device: Union[Type[Device], None], to_device_node_name: Union[str, None] = None):
        """
        This method sets the TO device and node of this connection.

        :param to_device: The TO device of this connection.
        :param to_device_node_name: The TO node of this connection (if it should be set, otherwise None).
        """
        if to_device is not None and isinstance(to_device, type) and not issubclass(to_device, Device):
            raise TypeError(f"detect illegal argument element {str(to_device)} for given attribute "
                            f"`to_device` - should be a subclasses of `balder.Device`")
        self._to_device = to_device

        if to_device_node_name is not None and not isinstance(to_device_node_name, str):
            raise TypeError(f"detect illegal argument type {type(to_device_node_name)} for given attribute "
                            f"`to_device_node_name` - should be a string value")
        self._to_device_node_name = to_device_node_name

    def get_conn_partner_of(self, device: Type[Device], node: Union[str, None] = None) -> Tuple[Type[Device], str]:
        """
        This method returns the connection partner of this connection - it always returns the other not given side

        :param device: the device itself - the other will be returned

        :param node: the node name of the device itself (only required if the connection starts and ends with the same
                     device)
        """
        if device not in (self.from_device, self.to_device):
            raise ValueError(f"the given device `{device.__qualname__}` is no component of this connection")
        if node is None:
            # check that the from_device and to_device are not the same
            if self.from_device == self.to_device:
                raise ValueError("the connection is a inner-device connection (start and end is the same device) - you "
                                 "have to provide the `node` string too")
            if device == self.from_device:
                return self.to_device, self.to_node_name

            return self.from_device, self.from_node_name

        if node not in (self.from_node_name, self.to_node_name):
            raise ValueError(f"the given node `{node}` is no component of this connection")

        if device == self.from_device and node == self.from_node_name:
            return self.to_device, self.to_node_name

        if device == self.to_device and node == self.to_node_name:
            return self.from_device, self.from_node_name

        raise ValueError(f"the given node `{node}` is no component of the given device `{device.__qualname__}`")

    def has_connection_from_to(
            self,
            start_device: Type[Device],
            start_device_node_name: Union[str, None] = None,
            end_device: Union[Type[Device], None] = None,
            end_device_node_name: Union[str, None] = None,
    ) -> bool:
        """
        This method checks if there is a connection from ``start_device`` to ``end_device``. This will return
        true if the ``start_device`` and ``end_device`` given in this method are also the ``start_device`` and
        ``end_device`` mentioned in this connection object. If this is a bidirectional connection, ``start_device`` and
        ``end_device`` can switch places.


        :param start_device: the device for which the method should check whether it is a communication partner (for
                             non-bidirectional connection, this has to be the start device)
        :param start_device_node_name: the node name that start device should have or None if it should be ignored
        :param end_device: the other device for which the method should check whether it is a communication partner (for
                           non-bidirectional connection, this has to be the end device - this is optional if only the
                           start device should be checked)
        :param end_device_node_name: the node name that start device should have or None if it should be ignored

        :return: returns true if the given direction is possible
        """
        if not (isinstance(start_device, type) and issubclass(start_device, Device)):
            raise TypeError("argument `start_device` needs to be a device")
        if not (start_device_node_name is None or isinstance(start_device_node_name, str)):
            raise TypeError("argument `start_device_node_name` needs to be a string or None if it should be ignored")
        if not (end_device is None or isinstance(end_device, type) and issubclass(end_device, Device)):
            raise TypeError("argument `end_device` needs to be a device or None if it should be ignored")
        if not (end_device_node_name is None or isinstance(end_device_node_name, str)):
            raise TypeError("argument `end_device_node_name` needs to be a string or None if it should be ignored")

        def check(start_dev, start_node_name, end_dev, end_node_name):
            if start_device != start_dev:
                return False
            if start_device_node_name is not None and start_device_node_name != start_node_name:
                return False
            if end_device is not None and end_device != end_dev:
                return False
            if end_device_node_name is not None and end_device_node_name != end_node_name:
                return False
            return True

        if self.bidirectional:
            if check(start_dev=self.to_device, start_node_name=self.to_node_name,
                     end_dev=self.from_device, end_node_name=self.from_node_name):
                return True
        return check(start_dev=self.from_device, start_node_name=self.from_node_name,
                     end_dev=self.to_device, end_node_name=self.to_node_name)

    def equal_with(self, other: ConnectionMetadata) -> bool:
        """
        This method returns true if the metadata of the current connection is equal with the metadata of the given
        connection.

        The method returns true in the following situations:
        * both connections are bidirectional and the from and to elements (device and node name) are the same
        * both connections are unidirectional and have the same from and to elements
        * both connections are bidirectional and the from is the to and the to is the from

        :return: true if the metadata of the current connection is contained in the metadata of the given one
        """
        return self.__compare_with(other, allow_single_unidirectional_for_both_directions=False)

    def contained_in(self, other: ConnectionMetadata) -> bool:
        """
        This method returns true if the metadata of the current connection is contained in the given one.

        The method returns true in the following situations:
        * both connections are bidirectional and the from and to elements (device and node name) are the same
        * both connections are unidirectional and have the same from and to elements
        * both connections are bidirectional and the from is the to and the to is the from
        * one connection is unidirectional and the other is bidirectional and the from and to elements are the same
        * one connection is unidirectional and the other is bidirectional and the from is the to and the to is the from

        :return: true if the metadata of the current connection is contained in the metadata of the given one
        """
        return self.__compare_with(other, allow_single_unidirectional_for_both_directions=True)

    @property
    def from_device(self):
        """device from which the connection starts"""
        return self._from_device

    @property
    def to_device(self):
        """device at which the connection ends"""
        return self._to_device

    @property
    def from_node_name(self):
        """the name of the node in the `Device` from which the connection starts"""
        return self._from_device_node_name

    @property
    def to_node_name(self):
        """the name of the node in the `Device` at which the connection ends"""
        return self._to_device_node_name

    @property
    def bidirectional(self) -> bool:
        """
        returns true if the connection is bidirectional (can go in both directions) otherwise false
        """
        return self._bidirectional



from __future__ import annotations
from typing import List
from enum import Enum


class FixtureDefinitionScope(Enum):
    """
    This enum describes the definition scope of a fixture. A definition scope is the position the fixture was defined.
    If the fixture was defined within the balderglob.py file, it has the definition-scope `GLOB`. If it is defined
    within a scenario or setup it has the equivalent SCENARIO or SETUP scope.
    """
    GLOB = 'glob'
    SETUP = 'setup'
    SCENARIO = 'scenario'

    @classmethod
    def get_order(cls) -> List[FixtureDefinitionScope]:
        """returns the priority order of the fixture definition scope"""
        return [cls.GLOB, cls.SETUP, cls.SCENARIO]



from _balder.testresult import ResultState
from _balder.balder_session import BalderSession
from ...test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0ValidSkipIgnoreTagsInScenario(Base0EnvtesterClass):
    """
    This testcase is a modified version of the envtester environment. It defines parent scenario classes that defines
    own IGNORE/SKIP marker. The test ensures that balder recognize them correctly and only resolves variations with
    scenarios that are valid to run.
    """

    @property
    def expected_data(self):
        return (
            # FIXTURE-CONSTRUCTION: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "construction"},
            [
                (
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_session
                    {"cls": "SetupA", "meth": "fixture_session", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: SetupB.fixture_session
                    {"cls": "SetupB", "meth": "fixture_session", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                )
            ],
            [
                (
                    # FIXTURE-CONSTRUCTION: ScenarioASub2.fixture_session
                    {"cls": "ScenarioAParent", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: ScenarioBSub1.fixture_session
                    {"cls": "ScenarioBParent", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: ScenarioBSub2.fixture_session
                    {"cls": "ScenarioBParent", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                )
            ],
            [
                (
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_setup
                    {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_setup
                    {"cls": "SetupA", "meth": "fixture_setup", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioASub2.fixture_setup
                    {"cls": "ScenarioAParent", "meth": "fixture_setup", "part": "construction"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
                    {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_scenario
                    {"cls": "SetupA", "meth": "fixture_scenario", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioASub2.fixture_scenario
                    {"cls": "ScenarioAParent", "meth": "fixture_scenario", "part": "construction"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
                    {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_variation
                    {"cls": "SetupA", "meth": "fixture_variation", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioASub2.fixture_variation
                    {"cls": "ScenarioAParent", "meth": "fixture_variation", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    [
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                             "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioASub2.fixture_testcase
                            {"cls": "ScenarioAParent", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # TESTCASE: ScenarioASub2.test_a_3
                            {"cls": "ScenarioAParent", "meth": "test_a_3"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: ScenarioASub2.fixture_testcase
                            {"cls": "ScenarioAParent", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                             "part": "teardown"},
                        ),
                    ],
                    # FIXTURE-TEARDOWN: ScenarioASub2.fixture_variation
                    {"cls": "ScenarioAParent", "meth": "fixture_variation", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupA.fixture_variation
                    {"cls": "SetupA", "meth": "fixture_variation", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_variation
                    {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},
                    # FIXTURE-TEARDOWN: ScenarioASub2.fixture_scenario
                    {"cls": "ScenarioAParent", "meth": "fixture_scenario", "part": "teardown"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupA.fixture_scenario
                    {"cls": "SetupA", "meth": "fixture_scenario", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_scenario
                    {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},
                    # FIXTURE-TEARDOWN: ScenarioASub2.fixture_setup
                    {"cls": "ScenarioAParent", "meth": "fixture_setup", "part": "teardown"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupA.fixture_setup
                    {"cls": "SetupA", "meth": "fixture_setup", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_setup
                    {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "teardown"},
                ),
                #
                # ------------------------------------------- SETUP CHANGE ---------------------------------------------
                #
                (
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_setup
                    {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupB.fixture_setup
                    {"cls": "SetupB", "meth": "fixture_setup", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    [
                        (
                            # FIXTURE-CONSTRUCTION: ScenarioBSub1.fixture_setup
                            {"cls": "ScenarioBParent", "meth": "fixture_setup", "part": "construction"},
                            {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                        ),
                        (
                            # FIXTURE-CONSTRUCTION: ScenarioBSub2.fixture_setup
                            {"cls": "ScenarioBParent", "meth": "fixture_setup", "part": "construction"},
                            {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                        )
                    ],
                    [
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
                            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupB.fixture_scenario
                            {"cls": "SetupB", "meth": "fixture_scenario", "part": "construction"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioBSub1.fixture_scenario
                            {"cls": "ScenarioBParent", "meth": "fixture_scenario", "part": "construction"},
                            {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
                            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupB.fixture_variation
                            {"cls": "SetupB", "meth": "fixture_variation", "part": "construction"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioBSub1.fixture_variation
                            {"cls": "ScenarioBParent", "meth": "fixture_variation", "part": "construction"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            [
                                (
                                    # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                                     "part": "construction"},
                                    # FIXTURE-CONSTRUCTION: SetupB.fixture_testcase
                                    {"cls": "SetupB", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-CONSTRUCTION: ScenarioBSub1.fixture_testcase
                                    {"cls": "ScenarioBParent", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                                    # TESTCASE: ScenarioBSub1.test_b_1
                                    {"cls": "ScenarioBParent", "meth": "test_b_1"},
                                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: ScenarioBSub1.fixture_testcase
                                    {"cls": "ScenarioBParent", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: SetupB.fixture_testcase
                                    {"cls": "SetupB", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                                     "part": "teardown"},
                                ),
                                (
                                    # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                                     "part": "construction"},
                                    # FIXTURE-CONSTRUCTION: SetupB.fixture_testcase
                                    {"cls": "SetupB", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-CONSTRUCTION: ScenarioBSub1.fixture_testcase
                                    {"cls": "ScenarioBParent", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                                    # TESTCASE: ScenarioBSub1.test_b_2
                                    {"cls": "ScenarioBParent", "meth": "test_b_2"},
                                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: ScenarioBSub1.fixture_testcase
                                    {"cls": "ScenarioBParent", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: SetupB.fixture_testcase
                                    {"cls": "SetupB", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                                     "part": "teardown"},
                                ),
                            ],
                            # FIXTURE-TEARDOWN: ScenarioBSub1.fixture_variation
                            {"cls": "ScenarioBParent", "meth": "fixture_variation", "part": "teardown"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupB.fixture_variation
                            {"cls": "SetupB", "meth": "fixture_variation", "part": "teardown"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_variation
                            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},
                            # FIXTURE-TEARDOWN: ScenarioBSub1.fixture_scenario
                            {"cls": "ScenarioBParent", "meth": "fixture_scenario", "part": "teardown"},
                            {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupB.fixture_scenario
                            {"cls": "SetupB", "meth": "fixture_scenario", "part": "teardown"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_scenario
                            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},
                        ),
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
                            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupB.fixture_scenario
                            {"cls": "SetupB", "meth": "fixture_scenario", "part": "construction"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioBSub2.fixture_scenario
                            {"cls": "ScenarioBParent", "meth": "fixture_scenario", "part": "construction"},
                            {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
                            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupB.fixture_variation
                            {"cls": "SetupB", "meth": "fixture_variation", "part": "construction"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioBSub2.fixture_variation
                            {"cls": "ScenarioBParent", "meth": "fixture_variation", "part": "construction"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            [
                                (
                                    # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                                     "part": "construction"},
                                    # FIXTURE-CONSTRUCTION: SetupB.fixture_testcase
                                    {"cls": "SetupB", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-CONSTRUCTION: ScenarioBSub2.fixture_testcase
                                    {"cls": "ScenarioBParent", "meth": "fixture_testcase", "part": "construction"},
                                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                                    # TESTCASE: ScenarioBSub2.test_b_3
                                    {"cls": "ScenarioBSub2", "meth": "test_b_3"},
                                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: ScenarioBSub2.fixture_testcase
                                    {"cls": "ScenarioBParent", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: SetupB.fixture_testcase
                                    {"cls": "SetupB", "meth": "fixture_testcase", "part": "teardown"},
                                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                                    # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                                    {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                                     "part": "teardown"},
                                ),
                            ],
                            # FIXTURE-TEARDOWN: ScenarioBSub2.fixture_variation
                            {"cls": "ScenarioBParent", "meth": "fixture_variation", "part": "teardown"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupB.fixture_variation
                            {"cls": "SetupB", "meth": "fixture_variation", "part": "teardown"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_variation
                            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},
                            # FIXTURE-TEARDOWN: ScenarioBSub2.fixture_scenario
                            {"cls": "ScenarioBParent", "meth": "fixture_scenario", "part": "teardown"},
                            {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupB.fixture_scenario
                            {"cls": "SetupB", "meth": "fixture_scenario", "part": "teardown"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_scenario
                            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},
                        )
                    ],
                    [
                        (
                            # FIXTURE-TEARDOWN: ScenarioBSub1.fixture_setup
                            {"cls": "ScenarioBParent", "meth": "fixture_setup", "part": "teardown"},
                            {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                        ),
                        (
                            # FIXTURE-TEARDOWN: ScenarioBSub2.fixture_setup
                            {"cls": "ScenarioBParent", "meth": "fixture_setup", "part": "teardown"},
                            {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                        )
                    ],
                    # FIXTURE-TEARDOWN: SetupB.fixture_setup
                    {"cls": "SetupB", "meth": "fixture_setup", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_setup
                    {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "teardown"},
                ),
            ],
            [
                (
                    # FIXTURE-TEARDOWN: ScenarioASub2.fixture_session
                    {"cls": "ScenarioAParent", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: ScenarioBSub1.fixture_session
                    {"cls": "ScenarioBParent", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: ScenarioBSub2.fixture_session
                    {"cls": "ScenarioBParent", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                )
            ],
            [
                (
                    # FIXTURE-TEARDOWN: SetupA.fixture_session
                    {"cls": "SetupA", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: SetupB.fixture_session
                    {"cls": "SetupB", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                )
            ],
            # FIXTURE-TEARDOWN: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "teardown"},
        )

    @staticmethod
    def validate_finished_session(session: BalderSession):

        expected_state = {
            'ScenarioASub1': {
                'test_a_1': ResultState.NOT_RUN,
                'test_a_2': ResultState.SKIP,
                'test_a_3': ResultState.SKIP,
            },
            'ScenarioASub2': {
                'test_a_1': ResultState.NOT_RUN,
                'test_a_2': ResultState.NOT_RUN,
                'test_a_3': ResultState.SUCCESS,
            },
            'ScenarioBSub1': {
                'test_b_1': ResultState.SUCCESS,
                'test_b_2': ResultState.SUCCESS,
                'test_b_3': ResultState.SKIP,
            },
            'ScenarioBSub2': {
                'test_b_1': ResultState.NOT_RUN,
                'test_b_2': ResultState.SKIP,
                'test_b_3': ResultState.SUCCESS,
            }
        }

        # check result states everywhere (have to be SUCCESS everywhere
        assert session.executor_tree.executor_result == ResultState.SUCCESS, \
            "test session does not terminates with success"

        assert session.executor_tree.construct_result.result == ResultState.SUCCESS, \
            "global executor tree construct part does not set ResultState.SUCCESS"
        assert session.executor_tree.body_result.result == ResultState.SUCCESS, \
            "global executor tree body part does not set ResultState.SUCCESS"
        assert session.executor_tree.teardown_result.result == ResultState.SUCCESS, \
            "global executor tree teardown part does not set ResultState.SUCCESS"

        for cur_setup_executor in session.executor_tree.get_setup_executors():
            assert cur_setup_executor.executor_result == ResultState.SUCCESS, \
                "the setup executor does not have result SUCCESS"

            assert cur_setup_executor.construct_result.result == ResultState.SUCCESS
            assert cur_setup_executor.body_result.result == ResultState.SUCCESS
            assert cur_setup_executor.teardown_result.result == ResultState.SUCCESS

            for cur_scenario_executor in cur_setup_executor.get_scenario_executors():
                expected_for_tests = expected_state[cur_scenario_executor.base_scenario_class.__class__.__name__]
                expected_branch_result = ResultState.NOT_RUN
                if ResultState.SUCCESS in expected_for_tests.values():
                    expected_branch_result = ResultState.SUCCESS
                elif ResultState.SKIP in expected_for_tests.values():
                    # has no success but has at least one SKIP
                    expected_branch_result = ResultState.SKIP

                assert cur_scenario_executor.executor_result == expected_branch_result, \
                    f"the scenario executor does not have result {expected_branch_result}"

                assert cur_scenario_executor.construct_result.result == \
                       (ResultState.SUCCESS if expected_branch_result == ResultState.SUCCESS
                        else ResultState.NOT_RUN)
                assert cur_scenario_executor.body_result.result == expected_branch_result
                assert cur_scenario_executor.teardown_result.result == \
                       (ResultState.SUCCESS if expected_branch_result == ResultState.SUCCESS
                        else ResultState.NOT_RUN)

                # expect same branch result for variation-executor, because we only have one per scenario
                for cur_variation_executor in cur_scenario_executor.get_variation_executors():
                    assert cur_variation_executor.executor_result == expected_branch_result, \
                        f"the variation executor does not have result {expected_branch_result}"

                    assert cur_variation_executor.construct_result.result == \
                       (ResultState.SUCCESS if expected_branch_result == ResultState.SUCCESS
                        else ResultState.NOT_RUN)
                    assert cur_variation_executor.body_result.result == expected_branch_result
                    assert cur_variation_executor.teardown_result.result == \
                       (ResultState.SUCCESS if expected_branch_result == ResultState.SUCCESS
                        else ResultState.NOT_RUN)

                    for cur_testcase_executor in cur_variation_executor.get_testcase_executors():
                        expected_test_result = expected_for_tests[cur_testcase_executor.base_testcase_callable.__name__]
                        assert cur_testcase_executor.executor_result == expected_test_result, \
                            f"the testcase executor does not have result {expected_test_result}"

                        assert cur_testcase_executor.construct_result.result == \
                               (ResultState.SUCCESS if expected_test_result == ResultState.SUCCESS
                                else ResultState.NOT_RUN)
                        assert cur_testcase_executor.body_result.result == expected_test_result
                        assert cur_testcase_executor.teardown_result.result == \
                               (ResultState.SUCCESS if expected_test_result == ResultState.SUCCESS
                                else ResultState.NOT_RUN)



from __future__ import annotations
from typing import Type, Union, List, TYPE_CHECKING

from _balder.fixture_execution_level import FixtureExecutionLevel
from _balder.testresult import ResultState, BranchBodyResult
from _balder.executor.basic_executable_executor import BasicExecutableExecutor
from _balder.executor.scenario_executor import ScenarioExecutor
from _balder.previous_executor_mark import PreviousExecutorMark
from _balder.controllers.setup_controller import SetupController
from _balder.controllers.device_controller import DeviceController
from _balder.controllers.feature_controller import FeatureController

if TYPE_CHECKING:
    from _balder.setup import Setup
    from _balder.scenario import Scenario
    from _balder.fixture_manager import FixtureManager
    from _balder.executor.executor_tree import ExecutorTree


class SetupExecutor(BasicExecutableExecutor):
    """
    A SetupExecutor is the highest branch object of an :class:`ExecutorTree`. It contains all scenarios and the
    underlying device mappings and their test cases that exist with this setup.
    """
    fixture_execution_level = FixtureExecutionLevel.SETUP

    def __init__(self, setup: Type[Setup], parent: ExecutorTree):
        super().__init__()
        self._scenario_executors: List[ScenarioExecutor] = []
        # check if instance already exists
        if hasattr(setup, "_instance") and setup._instance is not None and isinstance(setup._instance, setup):
            self._base_setup_class = setup._instance
        else:
            self._base_setup_class = setup()
            setup._instance = self._base_setup_class
        self._parent_executor = parent
        self._fixture_manager = parent.fixture_manager

        # contains the result object for the BODY part of this branch
        self.body_result = BranchBodyResult(self)

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    # ---------------------------------- CLASS METHODS ----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def all_child_executors(self) -> List[ScenarioExecutor]:
        return self._scenario_executors

    @property
    def parent_executor(self) -> ExecutorTree:
        return self._parent_executor

    @property
    def base_instance(self) -> object:
        """returns the base class instance to which this executor instance belongs"""
        return self.base_setup_class

    @property
    def base_setup_class(self) -> Setup:
        """returns the base :class:`Setup` that belongs to this executor"""
        return self._base_setup_class

    @property
    def base_setup_controller(self) -> SetupController:
        """returns the :class:`SetupController` for the setup object of this executor"""
        return SetupController.get_for(self.base_setup_class.__class__)

    @property
    def fixture_manager(self) -> FixtureManager:
        """returns the current active fixture manager for this executor"""
        return self._fixture_manager

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    def _prepare_execution(self, show_discarded):
        print(f"SETUP {self.base_setup_class.__class__.__name__}")

    def _body_execution(self, show_discarded):
        for cur_scenario_executor in self.get_scenario_executors(return_discarded=show_discarded):
            prev_mark = cur_scenario_executor.prev_mark
            if cur_scenario_executor.has_runnable_tests(consider_discarded_too=show_discarded) \
                    or cur_scenario_executor.has_skipped_tests():
                cur_scenario_executor.execute(show_discarded=show_discarded)
            elif prev_mark == PreviousExecutorMark.SKIP:
                cur_scenario_executor.set_result_for_whole_branch(ResultState.SKIP)
            elif prev_mark == PreviousExecutorMark.COVERED_BY:
                cur_scenario_executor.set_result_for_whole_branch(ResultState.COVERED_BY)
            else:
                cur_scenario_executor.set_result_for_whole_branch(ResultState.NOT_RUN)

    def _cleanup_execution(self, show_discarded):
        pass

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def update_inner_referenced_feature_instances(self):
        """
        This method ensures that all inner referenced feature instances of all devices that are used in this setup are
        set to the correct feature instance. For every existing device, this method goes trough all assigned features
        and checks for a inner-referenced feature. It makes sure, that every inner-referenced feature variable has the
        final assigned setup feature that belongs to it.

        # TODO check where we validate that inner references feature exists in setup
        """
        for cur_setup_device in self.base_setup_controller.get_all_abs_inner_device_classes():
            # these features are subclasses of the real defined one (because they are already the replaced ones)
            all_device_features = DeviceController.get_for(cur_setup_device).get_all_instantiated_feature_objects()
            all_instantiated_feature_objs_of_this_dev = [cur_feature for _, cur_feature in all_device_features.items()]
            for _, cur_feature in all_device_features.items():
                cur_feature_controller = FeatureController.get_for(cur_feature.__class__)
                # now check the inner referenced features of this feature and check if that exists in the device
                for cur_ref_feature_name, cur_ref_feature in \
                        cur_feature_controller.get_inner_referenced_features().items():
                    potential_candidates = [
                        candidate_feature for candidate_feature in all_instantiated_feature_objs_of_this_dev
                        if isinstance(candidate_feature, cur_ref_feature.__class__)]
                    if len(potential_candidates) != 1:
                        raise RuntimeError("found none or more than one potential replacing candidates")
                    replacing_candidate = potential_candidates[0]
                    # because `cur_feature` is only the object instance, the value will be overwritten only for this
                    # object
                    setattr(cur_feature, cur_ref_feature_name, replacing_candidate)

    def get_scenario_executors(self, return_discarded=False) -> List[ScenarioExecutor]:
        """
        returns a list with all scenario executors that belongs to this setup executor

        :param return_discarded: True if the method should return discarded variations too

        :return: a list of relevant :class:`ScenarioExecutor`
        """
        if return_discarded:
            return self._scenario_executors

        return [
            cur_scenario_executor
            for cur_scenario_executor in self._scenario_executors
            if len(cur_scenario_executor.get_variation_executors(return_discarded=False)) > 0
        ]

    def cleanup_empty_executor_branches(self, consider_discarded=False):
        to_remove_executor = []
        for cur_scenario_executor in self.get_scenario_executors(return_discarded=consider_discarded):
            cur_scenario_executor.cleanup_empty_executor_branches(consider_discarded=consider_discarded)
            if len(cur_scenario_executor.get_variation_executors(return_discarded=consider_discarded)) == 0:
                # remove this whole executor because it has no children anymore
                to_remove_executor.append(cur_scenario_executor)
        for cur_scenario_executor in to_remove_executor:
            self._scenario_executors.remove(cur_scenario_executor)

    def add_scenario_executor(self, scenario_executor: ScenarioExecutor):
        """
        This method adds a new ScenarioExecutor to the child element list of the tree
        """
        if not isinstance(scenario_executor, ScenarioExecutor):
            raise TypeError("the given object `scenario_executor` must be of type type `ScenarioExecutor`")
        if scenario_executor in self._scenario_executors:
            raise ValueError("the given object `scenario_executor` already exists in child list")
        self._scenario_executors.append(scenario_executor)

    def get_executor_for_scenario(self, scenario: Type[Scenario]) -> Union[ScenarioExecutor, None]:
        """
        This method searches for a ScenarioExecutor in the internal list for which the given scenario is
        contained in

        :param scenario: the scenario class for which the executor should be searched for

        :return: returns the associated ScenarioExecutor or None if no matching could be found
        """
        for cur_scenario_executor in self._scenario_executors:
            if cur_scenario_executor.base_scenario_class.__class__ == scenario:
                return cur_scenario_executor
        # can not find some
        return None



from __future__ import annotations
from typing import List, Tuple, Union, Type, Dict

import copy
import itertools

from _balder.connection_metadata import ConnectionMetadata
from _balder.device import Device
from _balder.exceptions import IllegalConnectionTypeError
from _balder.cnnrelations import AndConnectionRelation, OrConnectionRelation
from _balder.utils.functions import cnn_type_check_and_convert


class ConnectionType(type):
    """
    Metaclass for :class:`Connection` objects
    """

    def __and__(cls, other) -> AndConnectionRelation:
        return cls().__and__(other)

    def __or__(cls, other) -> OrConnectionRelation:
        return cls().__or__(other)


class Connection(metaclass=ConnectionType):
    """
    This is the basic connection class. On one side it is the common base class for all connection objects. On the other
    side it can also be used to describe a container connection, that contains a list of different connection items.
    """
    #: contains all parent connection information for every existing connection
    __parents: Dict[Type[Connection], Dict[str, List[Type[Connection]]]] = {}

    def __init__(self, from_device: Union[Type[Device], None] = None, to_device: Union[Type[Device], None] = None,
                 from_device_node_name: Union[str, None] = None, to_device_node_name: Union[str, None] = None):
        """
        This method creates a new Connection-Tree.

        .. note::
            With a direct instance of a :class:`Connection` object you can create an own Connection-Tree. You can use
            this container object, if you need a container for a list of :class:`Connection`-Trees that are combined
            with an AND (``&``) or/and an OR (``|``).

        :param from_device: the device this connection starts from (default: None)

        :param to_device: the device this connection ends in (default: None)

        :param from_device_node_name: the node name of the device the connection start

        :param to_device_node_name: the node name of the device the connection ends
        """
        # note: currently every connection is bidirectional (we want to add support for this later)

        # contains all metadata of this connection object
        self._metadata = ConnectionMetadata(
            from_device=from_device,
            to_device=to_device,
            from_device_node_name=from_device_node_name,
            to_device_node_name=to_device_node_name,
            bidirectional=True
        )

        # contains all sub connection objects, this connection tree is based on
        self._based_on_connections = OrConnectionRelation()

    def __and__(self, other: Union[Connection, AndConnectionRelation, OrConnectionRelation]) -> AndConnectionRelation:
        new_list = AndConnectionRelation(self)

        other = cnn_type_check_and_convert(other)
        new_list.append(other)

        return new_list

    def __or__(self, other: Union[Connection, AndConnectionRelation, OrConnectionRelation]) -> OrConnectionRelation:
        new_list = OrConnectionRelation(self)

        other = cnn_type_check_and_convert(other)
        new_list.append(other)

        return new_list

    def __eq__(self, other):
        if isinstance(other, Connection):
            return self.equal_with(other)

        return False

    def __hash__(self):
        all_hashes = hash(self.metadata) + hash(self.based_on_elements)
        return hash(all_hashes)

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    # ---------------------------------- CLASS METHODS -----------------------------------------------------------------

    @classmethod
    def get_parents(cls, tree_name: Union[str, None] = None) -> List[Type[Connection]]:
        """
        This method returns the parent classes of this connection.

        :param tree_name: the tree name the parents should be returned (default: use tree defined in `GlobalSetting`)
        """
        from _balder.balder_session import BalderSession  # pylint: disable=import-outside-toplevel
        if tree_name is None:
            tree_name = BalderSession.get_current_active_global_conntree_name()
        conn_dict = Connection.__parents.get(cls, {})
        parent_list = conn_dict.get(tree_name)
        return [] if parent_list is None else parent_list

    @classmethod
    def set_parents(cls, data: Union[List[Type[Connection]], None], tree_name: str = ""):
        """
        This method allows to set the data for the parent classes of this connection - note that None will remove the
        entry and completely disconnect this object from the global connection tree

        :param data: the new parents (None if everything should be reset)

        :param tree_name: the tree name of the parents that should be set (default: global tree)
        """
        if data is None:
            if cls in Connection.__parents.keys():
                if tree_name in Connection.__parents[cls].keys():
                    del Connection.__parents[cls][tree_name]
        else:
            if Connection.__parents.get(cls) is None:
                Connection.__parents[cls] = {}
            Connection.__parents[cls][tree_name] = data

    @classmethod
    def is_child_of(cls: Type[Connection], other_conn: Type[Connection]) -> bool:
        """
        determines whether this connection is a child of the given connection
        """
        return other_conn.is_parent_of(cls)

    @classmethod
    def is_parent_of(cls: Type[Connection], other_conn: Type[Connection]) -> bool:
        """
        determines whether this connection is a parent of the given connection
        """
        if other_conn.get_parents():
            if cls in other_conn.get_parents():
                return True
            for cur_higher_other_child in other_conn.get_parents():
                # ignore tuples, because this will return False (check with tuples not possible)
                if not isinstance(cur_higher_other_child, tuple):
                    if cls.is_parent_of(other_conn=cur_higher_other_child):
                        return True
            return False
        # the other connection has no parents, so this can not be the parent class
        return False

    @classmethod
    def based_on(
            cls,
            connection: Union[Connection, Type[Connection], AndConnectionRelation, OrConnectionRelation]
    ) -> Connection:
        """
        With this method it is possible to define several sublayers of the connection. You can pass various other
        connections in this method as arguments.

        Note that you can define logical statements with the `|` and the `&` operator.
        For example: ``Connection1() or Connection2() and Connection3()``.

        :param connection: all connection items for which this connection is based on

        """
        this_instance = cls()

        if isinstance(connection, type) and issubclass(connection, Connection):
            connection = connection()
        new_items = []
        if isinstance(connection, Connection):
            new_items.append(connection)
        elif isinstance(connection, AndConnectionRelation):
            new_items.append(connection)
        elif isinstance(connection, OrConnectionRelation):
            for cur_inner_elem in connection.connections:
                new_items.append(cur_inner_elem)
        else:
            raise TypeError(f'can not use object from type {connection}')

        # do not create a container connection if no container is required here
        if cls == Connection and len(new_items) == 1 and isinstance(new_items[0], Connection):
            return new_items[0]
        this_instance.append_to_based_on(*new_items)
        return this_instance

    @classmethod
    def filter_connections_that_are_contained_in(
            cls,
            cnns_from: List[Union[Connection, AndConnectionRelation, OrConnectionRelation]],
            are_contained_in: Connection,
            ignore_metadata: bool = False
    ) -> List[Connection]:
        """
        This method filters the connection elements from the first list to include only those connections that are
        contained within the provided connection ``are_contained_in``.

        :param cnns_from: a list of connections
        :param are_contained_in: the connection, the connection elements should be contained in
        :param ignore_metadata: True, if the metadata should be ignored
        :return: a list with the filtered connections
        """
        return [
            cnn for cnn in cnns_from
            if cnn.contained_in(other_conn=are_contained_in, ignore_metadata=ignore_metadata)
        ]

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def metadata(self) -> ConnectionMetadata:
        """returns the connection metadata dictionary"""
        return self._metadata

    @property
    def from_device(self):
        """device from which the connection starts"""
        return self._metadata.from_device if self._metadata else None

    @property
    def to_device(self):
        """device at which the connection ends"""
        return self._metadata.to_device if self._metadata else None

    @property
    def from_node_name(self):
        """the name of the node in the `Device` from which the connection starts"""
        return self._metadata.from_node_name if self._metadata else None

    @property
    def to_node_name(self):
        """the name of the node in the `Device` at which the connection ends"""
        return self._metadata.to_node_name if self._metadata else None

    @property
    def based_on_elements(self) -> OrConnectionRelation:
        """returns a copy of the internal based_on_connection"""
        return self._based_on_connections.clone()

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    def _is_directly_contained_in(self, other_conn: Connection, ignore_metadata: bool) -> bool:
        """
        Helper method that returns true if this connection is directly contained in the `other_conn`.

        .. note::
            'Directly contained-in' means that the resolved versions of both connections are of the same connection
            type, and the elements of that connection are directly embedded within one another from that point onward.
            It does not check deeper into the other_conn.

        :param other_conn: the other connection
        :param ignore_metadata: True, if the metadata should be ignored
        """

        resolved_self = self.get_resolved()
        resolved_other = other_conn.get_resolved()

        if resolved_self.__class__ == Connection and len(resolved_self.based_on_elements) == 0 or \
                resolved_other.__class__ == Connection and len(resolved_other.based_on_elements) == 0:
            # one of the resolved object is a raw `Connection` object without based-on-elements -> always true
            return True

        if resolved_self.__class__ == Connection:
            return resolved_self.based_on_elements.contained_in(resolved_other, ignore_metadata=ignore_metadata)

        if resolved_self.__class__ == resolved_other.__class__:
            # The element itself has already matched, now we still have to check whether at least one inner element
            # of this type is contained in minimum one element of the other
            singles_self = resolved_self.get_singles()
            singles_other = other_conn.get_singles()

            # check that for one single_self element all hierarchical based_on elements are in one of the single
            # other element
            for cur_single_self, cur_single_other in itertools.product(singles_self, singles_other):
                # check if both consists of only one element
                if len(cur_single_self.based_on_elements) == 0:
                    # the cur self single is only one element -> this is contained in the other
                    return True

                if len(cur_single_other.based_on_elements) == 0:
                    # the other element is only one element, but the self element not -> contained_in
                    # for this single definitely false
                    continue

                # note: for both only one `based_on_elements` is possible, because they are singles
                self_first_basedon = cur_single_self.based_on_elements[0]
                other_first_basedon = cur_single_other.based_on_elements[0]

                if isinstance(self_first_basedon, Connection) and \
                        isinstance(other_first_basedon, AndConnectionRelation):
                    continue

                # find a complete valid match
                if self_first_basedon.contained_in(other_first_basedon, ignore_metadata=ignore_metadata):
                    return True
        return False

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def get_intersection_with_other_single(
            self,
            other_conn: Union[Connection, AndConnectionRelation, OrConnectionRelation]
    ) -> List[Connection]:
        """
        A helper method that returns an intersection between the two connections (self and the given one).

        :param other_conn: the other **single** connection object (could be a relation too, but note that them needs to
                           be **single** elements!)
        """
        if not self.is_single():
            raise ValueError(
                "the current connection object is not single -> method only possible for single connections")
        if not other_conn.is_single():
            raise ValueError(
                "the other connection object is not single -> method only possible for single connections")

        intersection = []

        #: check if some sub elements of self connection are contained in `other_conn`
        self_pieces = self.cut_into_all_possible_pieces()
        intersection.extend(self.__class__.filter_connections_that_are_contained_in(
            cnns_from=self_pieces,
            are_contained_in=other_conn,
            ignore_metadata=True
        ))

        #: check if some sub elements of `other_conn` are contained in self connection
        other_pieces = other_conn.cut_into_all_possible_pieces()
        intersection.extend(self.__class__.filter_connections_that_are_contained_in(
            cnns_from=other_pieces,
            are_contained_in=self,
            ignore_metadata=True
        ))

        #: filter all duplicated (and contained in each other) connections
        intersection_without_duplicates = []
        for cur_conn in intersection:
            found_it = False
            for cur_existing_conn in intersection_without_duplicates:
                if cur_conn.equal_with(cur_existing_conn, ignore_metadata=True):
                    found_it = True
                    break
            if not found_it:
                intersection_without_duplicates.append(cur_conn)

        intersection_filtered = []
        #: filter all *contained in each other* connections
        for cur_conn in intersection_without_duplicates:
            is_contained_in_another = False
            for cur_validate_cnn in intersection_without_duplicates:
                if cur_validate_cnn == cur_conn:
                    # skip the same element
                    continue
                if cur_conn.contained_in(cur_validate_cnn, ignore_metadata=True):
                    is_contained_in_another = True
                    break
            if not is_contained_in_another:
                if isinstance(cur_conn, (AndConnectionRelation,OrConnectionRelation)):
                    cur_conn = Connection.based_on(cur_conn)
                intersection_filtered.append(cur_conn)

        # only return unique objects
        return intersection_filtered

    def clone_without_based_on_elements(self) -> Connection:
        """
        This method returns a copied version of this element, while all `_based_on_connections` are removed (the copied
        element has an empty list here).

        :return: a python copied object of this item
        """
        self_copy = copy.copy(self)
        self_copy._based_on_connections = OrConnectionRelation()  # pylint: disable=protected-access
        return self_copy

    def clone(self) -> Connection:
        """
        This method returns an exact clone of this connection. For this clone every inner connection object will be
        newly instantiated, but all internal references (like the devices and so on) will not be copied (objects are the
        same for this object and the clone). The method will make a normal copy for every connection object in the
        `_based_on_elements` list.
        """
        self_copy = self.clone_without_based_on_elements()
        self_copy.append_to_based_on(self._based_on_connections.clone())  # pylint: disable=protected-access
        return self_copy

    def cut_into_all_possible_subtree_branches(self) -> List[Connection]:
        """
        This method returns a list of all possible connection tree branches. A branch is a single connection, while
        this method returns a list of all possible singles where every single connection has this connection as head.
        """
        all_pieces = [self.clone()]
        if len(self.based_on_elements) == 0:
            return all_pieces

        if not self.is_single():
            raise ValueError("the current connection object is not single -> method only possible for single "
                             "connections")

        # return all possibilities of the relation while remain the current object as head
        for sub_branch in self.based_on_elements.cut_into_all_possible_subtree_branches():
            copied_conn = self.clone_without_based_on_elements()
            copied_conn.append_to_based_on(sub_branch)
            all_pieces.append(copied_conn)
        return all_pieces

    def cut_into_all_possible_pieces(self) -> List[Connection]:
        """
        This method cuts the resolved connection tree in all possible pieces by removing elements that change
        the existing tree - thereby the method returns a list with all possibilities (a copy of this object with all
        elements is contained too).

        .. note::
            Note that the given element or the child elements of a given direct :class:`Connection` object has to be
            single!
        """
        if self.__class__ == Connection:
            # this is only a container, execute process for every item of this one
            child_elems = self.based_on_elements.connections
        else:
            child_elems = [self]

        all_pieces = []

        for cur_item in child_elems:

            if not cur_item.is_single():
                raise ValueError("one of the given element is not single -> method only works with single items")

            cur_element = cur_item
            while cur_element:

                if isinstance(cur_element, AndConnectionRelation):
                    # now we can not go deeper, because we need this current AND connection
                    all_pieces += [Connection.based_on(and_relation)
                                   for and_relation in cur_element.cut_into_all_possible_subtree_branches()]
                    break

                all_pieces += cur_element.cut_into_all_possible_subtree_branches()

                if len(cur_element.based_on_elements.connections) > 1:
                    # should never be fulfilled because we have a single connection
                    raise ValueError('unexpected size of inner connections')

                cur_element = cur_element.based_on_elements.connections[0] \
                    if cur_element.based_on_elements.connections else None

        # filter all duplicates
        return list(set(all_pieces))

    def set_metadata_for_all_subitems(self, metadata: Union[ConnectionMetadata, None]):
        """
        This method sets the metadata for all existing Connection items in this element.

        :param metadata: the metadata that should be set (if the value is explicitly set to None, it removes the
                         metadata from every item)
        """
        if metadata is None:
            metadata = ConnectionMetadata()

        if not isinstance(metadata, ConnectionMetadata):
            raise TypeError('metadata must be an instance of `ConnectionMetadata`')

        for cur_base_elem in self._based_on_connections:
            cur_base_elem.set_metadata_for_all_subitems(metadata=metadata)
        self._metadata = metadata

    def get_tree_str(self) -> str:
        """
        This method returns a string, that represents all internal connections as readable string.

        :return: a readable string of the whole connection tree
        """
        if len(self.based_on_elements) == 0:
            return f"{self.__class__.__name__}()"

        based_on_strings = []
        for cur_elem in self.based_on_elements:
            if isinstance(cur_elem, tuple):
                based_on_strings.append(
                    f"({', '.join([cur_tuple_elem.get_tree_str() for cur_tuple_elem in cur_elem])})")
            else:
                based_on_strings.append(cur_elem.get_tree_str())

        return f"{self.__class__.__name__}.based_on({'| '.join(based_on_strings)})"

    def is_bidirectional(self) -> Union[bool, None]:
        """
        Provides the information in which direction the connection is supported (if the property returns true, the
        connection must work in both directions, otherwise it is a unidirectional connection from the `from_device` to
        the` to_device`
        """
        return self._metadata.bidirectional if self._metadata else None

    def is_universal(self):
        """
        Provides the information if the current connection object is a universal connection. This means, that the type
        is the base :class:`Connection` and the based_on_elements are empty.
        """
        return self.__class__ == Connection and len(self._based_on_connections) == 0

    def is_resolved(self):
        """
        This method returns true if the given (sub-)connection tree is completely resolved (describes every single
        Connection object of the tree - there are no undefined connection layers between an object and the given
        parent).
        """
        if self.__class__ == Connection:
            return self.based_on_elements.is_resolved()

        self_parents = self.__class__.get_parents()

        for cur_based_on in self.based_on_elements.connections:
            if isinstance(cur_based_on, Connection):
                if cur_based_on.__class__ not in self_parents:
                    return False
            elif isinstance(cur_based_on, (AndConnectionRelation, OrConnectionRelation)):
                for cur_type in cur_based_on.get_all_used_connection_types():
                    if cur_type not in self_parents:
                        return False
            if not cur_based_on.is_resolved():
                return False
        return True

    def is_single(self):
        """
        This method returns true if there exists no logical **OR** in the based on connection(s). One **AND** relation
        is allowed.

        .. note::
            Note that this method also returns False, if the connection is not completely resolved!
        """
        if len(self.based_on_elements) == 0:
            # tree ends here
            return True

        if len(self.based_on_elements) > 1:
            # more than one element -> not single
            return False

        if not self.is_resolved():
            return False

        return self.based_on_elements[0].is_single()

    def get_resolved(self) -> Connection:
        """
        This method returns a resolved Connection Tree. This means that it convert the based_on references so that
        every based on connection is a direct parent of the current element. It secures that there are no undefined
        connection layers between an object and the given parent.

        .. note::
            This method returns the connection without a container :class:`Connection`, if the container
            :class:`Connection` would only consist of one single connection (which is no AND relation!). In that case
            the method returns this child connection directly without any :class:`Connection` container otherwise the
            :class:`Connection` container class with all resolved child classes will be returned.
        """
        copied_base = self.clone_without_based_on_elements()

        if self.is_resolved():
            copied_base.append_to_based_on(self.based_on_elements.get_simplified_relation())
        elif self.__class__ == Connection:
            # the base object is a container Connection - iterate over the items and determine the values for them
            copied_base.append_to_based_on(self.based_on_elements.get_simplified_relation().get_resolved())
        else:
            # independent which based-on elements we have, we need to determine all elements between this connection
            # and the elements of the relation
            simplified_based_on = self.based_on_elements.get_simplified_relation()

            for next_higher_parent in simplified_based_on:
                if isinstance(next_higher_parent, AndConnectionRelation):
                    # determine all possibilities
                    for new_and_relation in next_higher_parent.get_possibilities_for_direct_parent_cnn(self.__class__):
                        copied_base.append_to_based_on(new_and_relation)
                else:
                    # `next_higher_parent` needs to be a connection, because we are using simplified which has only
                    # `OR[AND[Cnn, ...], Cnn, ..]`
                    if next_higher_parent.__class__ in self.__class__.get_parents():
                        # is already a direct parent
                        copied_base.append_to_based_on(next_higher_parent.get_resolved())
                        continue
                    # only add the first level of direct parents - deeper will be added by recursively call of
                    # `get_resolved`
                    for cur_self_direct_parent in self.__class__.get_parents():
                        if next_higher_parent.__class__.is_parent_of(cur_self_direct_parent):
                            new_child = cur_self_direct_parent.based_on(next_higher_parent)
                            copied_base.append_to_based_on(new_child.get_resolved())

        # if it is a connection container, where only one element exists that is no AND relation -> return this directly
        # instead of the container
        if copied_base.__class__ == Connection and len(copied_base.based_on_elements) == 1 and not \
                isinstance(copied_base.based_on_elements[0], AndConnectionRelation):
            return copied_base.based_on_elements[0]

        return copied_base

    def get_singles(self) -> List[Connection]:
        """
        This method returns a list with connection objects where every element is a single clear possibility of the
        current connection sub-tree. With this method the **OR** possibilities will be resolved, so that every item does
        not contain any OR relations.

        .. note::
            If the current object is a container :class:`Connection` object (direct instance of :class:`Connection`),
            this method returns the single elements without a container class!
        """
        # note: This method always work with the resolved and simplified version of the object because it is using
        # `get_resolved()`.
        if self.is_universal():
            return [self]

        if self.is_single():
            return [self]

        all_singles = []

        resolved_self = self.get_resolved()

        for cur_child in resolved_self.based_on_elements:
            for cur_single_child in cur_child.get_singles():
                cur_single_child = cur_single_child.clone()
                if self.__class__ == Connection and isinstance(cur_single_child, Connection):
                    all_singles.append(cur_single_child)
                else:
                    copied_base = resolved_self.clone_without_based_on_elements()
                    copied_base.append_to_based_on(cur_single_child)
                    all_singles.append(copied_base)

        return all_singles

    def get_conn_partner_of(self, device: Type[Device], node: Union[str, None] = None) -> Tuple[Type[Device], str]:
        """
        This method returns the connection partner of this connection - it always returns the other not given side

        :param device: the device itself - the other will be returned

        :param node: the node name of the device itself (only required if the connection starts and ends with the same
                     device)
        """
        return self.metadata.get_conn_partner_of(device, node)

    def has_connection_from_to(
            self,
            start_device: Type[Device],
            start_device_node_name: Union[str, None] = None,
            end_device: Union[Type[Device], None] = None,
            end_device_node_name: Union[str, None] = None,
    ) -> bool:
        """
        This method checks if there is a connection from ``start_device`` to ``end_device``. This will return
        true if the ``start_device`` and ``end_device`` given in this method are also the ``start_device`` and
        ``end_device`` mentioned in this connection object. If this is a bidirectional connection, ``start_device`` and
        ``end_device`` can switch places.


        :param start_device: the device for which the method should check whether it is a communication partner (for
                             non-bidirectional connection, this has to be the start device)
        :param start_device_node_name: the node name that start device should have or None if it should be ignored
        :param end_device: the other device for which the method should check whether it is a communication partner (for
                           non-bidirectional connection, this has to be the end device - this is optional if only the
                           start device should be checked)
        :param end_device_node_name: the node name that start device should have or None if it should be ignored

        :return: returns true if the given direction is possible
        """
        return self.metadata.has_connection_from_to(
            start_device=start_device,
            start_device_node_name=start_device_node_name,
            end_device=end_device,
            end_device_node_name=end_device_node_name
        )

    def equal_with(self, other_conn: Connection, ignore_metadata=False) -> bool:
        """
        This method returns True if the current object matches the given object. It always converts the elements to a
        resolved version and checks if both of them are exactly the same.

        .. note::
            Note that both Connection objects have to have the same ending parent elements. Only the order of the parent
            elements are obviously irrelevant.

        .. note::
            Note that it doesn't matter if the connection is embedded in a container-Connection element (direct
            instance of :class`Connection`) or not. It only checks, that the logical data of them are the same.

        :param other_conn: the other connection, this connection will be compared with

        :param ignore_metadata: if this value is true the method ignores the metadata

        :return: returns True if both elements are same
        """
        if not ignore_metadata:
            metadata_check_result = self.metadata.equal_with(other_conn.metadata)
            if metadata_check_result is False:
                return False

        # we do not need to check something for the container :class:`Connection`, because the `.get_resolved()` will
        # return no container if it only consists of one child connection
        resolved_self = self.get_resolved()
        resolved_other = other_conn.get_resolved()

        if self.__class__ != other_conn.__class__:
            return False

        return resolved_self.based_on_elements.equal_with(resolved_other.based_on_elements)

    def contained_in(
            self,
            other_conn: Connection | AndConnectionRelation | OrConnectionRelation,
            ignore_metadata=False
    ) -> bool:
        """
        This method determines if one connection tree can be embedded within another connection tree. A connection
        object represents a specific segment of the extensive connection tree managed by Balder. The method evaluates
        whether at least one SINGLE connection object of this connection tree fits within a possible SINGLE connection
        of the provided connection `other_conn`.

        .. note::
            The method returns true if one single connection of this object fits in another single connection that is
            given by `other_conn`.

        :param other_conn: the other connection

        :param ignore_metadata: if this value is true the method ignores the metadata

        :return: true if the self object is contained in the `other_conn`, otherwise false
        """
        # note: This method always work with the resolved and simplified version of the object because it is using
        # `get_resolved()`.

        if not ignore_metadata:
            metadata_check_result = self.metadata.contained_in(other_conn.metadata)
            if metadata_check_result is False:
                return False

        resolved_self = self.get_resolved()
        resolved_other = other_conn.get_resolved()

        if self._is_directly_contained_in(resolved_other, ignore_metadata=ignore_metadata):
            return True

        # the elements itself do not match -> go deeper within the other connection

        resolved_other_relation = resolved_other.based_on_elements \
            if isinstance(resolved_other, Connection) else resolved_other

        for cur_other_based_on in resolved_other_relation.connections:
            # `cur_other_based_on` can only be a Connection or an AND (resolved can not ba a inner OR)
            if isinstance(cur_other_based_on, AndConnectionRelation):
                # check if the current connection fits in one of the AND relation items -> allowed too (f.e. a
                # smaller AND contained in a bigger AND)
                for cur_other_and_element in cur_other_based_on.connections:
                    if resolved_self.contained_in(cur_other_and_element, ignore_metadata=ignore_metadata):
                        return True
            else:
                if resolved_self.contained_in(cur_other_based_on, ignore_metadata=ignore_metadata):
                    # element was found in this branch
                    return True
        return False

    def intersection_with(
            self, other_conn: Union[Connection, Type[Connection], AndConnectionRelation, OrConnectionRelation]) \
            -> Union[Connection, None]:
        """
        This method returns a list of subtrees that describe the intersection of this connection subtree and the
        given ones. Note that this method converts all connections in **single** **resolved** connections first.
        The method checks if there are common intersections between the elements of this object and the given connection
        elements within the single connections.
        The method cleans up this list and only return unique sub-connection trees!

        :param other_conn: the other sub connection tree list

        :return: the intersection connection or none, if the method has no intersection
        """
        other_conn = cnn_type_check_and_convert(other_conn)

        if isinstance(other_conn, Connection) and other_conn.__class__ == Connection:
            if len(other_conn.based_on_elements) == 0:
                return self.clone()
            other_conn = other_conn.based_on_elements

        if self.is_universal():
            return other_conn.clone() if isinstance(other_conn, Connection) else Connection.based_on(other_conn.clone())

        # determine all single connection of the two sides (could contain AND relations, where every element is a single
        # connection too)
        self_conn_singles = self.get_singles()
        other_conn_singles = other_conn.get_singles()

        intersections = []
        # determine intersections between all of these single components
        for cur_self_conn, cur_other_conn in itertools.product(self_conn_singles, other_conn_singles):
            for cur_intersection in cur_self_conn.get_intersection_with_other_single(cur_other_conn):
                intersections.append(cur_intersection)

        intersections = set(intersections)

        #: filter all *contained in each other* connections
        intersection_filtered = []
        for cur_conn in intersections:
            is_contained_in_another = False
            for cur_validate_cnn in intersections:
                if cur_validate_cnn == cur_conn:
                    # skip the same element
                    continue
                if cur_conn.contained_in(cur_validate_cnn, ignore_metadata=True):
                    is_contained_in_another = True
                    break
            if not is_contained_in_another:
                intersection_filtered.append(cur_conn)

        if len(intersection_filtered) == 0:
            # there is no intersection
            return None

        return Connection.based_on(OrConnectionRelation(*[cnn.clone() for cnn in intersection_filtered]))

    def append_to_based_on(
            self, *args: Union[Type[Connection], Connection, OrConnectionRelation, AndConnectionRelation]) -> None:
        """
        with this method you can extend the internal based_on list with the transferred elements. Any number of
        :meth:`Connection` objects or relations with :meth:`Connection` objects can be given to this method. They will
        all be added to the internal OR relation.

        :param args: all connection items that should be added here
        """

        def validate_that_subconnection_is_parent(idx, connection_type: Type[Connection]):
            if connection_type == Connection:
                raise ValueError(f"it is not allowed to provide a container Connection object in based_on items - "
                                 f"found at index {idx}")
            # `based_on` call for this sub-connection because we get an `Connection` object
            if self.__class__ != Connection:
                if not connection_type.is_parent_of(self.__class__):
                    raise IllegalConnectionTypeError(
                        f"the connection `{cur_elem.__class__.__name__}` (at parameter pos {idx}) is "
                        f"no parent class of the `{self.__class__.__name__}`")

        for cur_idx, cur_elem in enumerate(args):
            if isinstance(cur_elem, type) and issubclass(cur_elem, Connection):
                cur_elem = cur_elem()

            if isinstance(cur_elem, Connection):
                if cur_elem.is_universal():
                    # ignore it, because it is irrelevant for this connection
                    continue
                validate_that_subconnection_is_parent(cur_idx, cur_elem.__class__)
                self._based_on_connections.append(cur_elem)
            elif isinstance(cur_elem, OrConnectionRelation):
                for cur_inner_elem in cur_elem.connections:
                    if isinstance(cur_inner_elem, Connection) and cur_inner_elem.is_universal():
                        # ignore it, because it is irrelevant for this connection
                        continue
                    validate_that_subconnection_is_parent(cur_idx, cur_inner_elem.__class__)
                    self._based_on_connections.append(cur_inner_elem)
            elif isinstance(cur_elem, AndConnectionRelation):
                for cur_inner_type in cur_elem.get_all_used_connection_types():
                    validate_that_subconnection_is_parent(cur_idx, cur_inner_type)
                self._based_on_connections.append(cur_elem)
            else:
                raise TypeError(f"illegal type `{cur_elem.__name__}` for parameter at pos {cur_idx}")



from ...lib.features import FeatureIV
from ...balderglob import RuntimeObserver


class SetupFeatureIV(FeatureIV):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureIV, SetupFeatureIV.do_something, "enter `SetupFeatureIV.do_something`",
            category="feature")



from ...lib.features import FeatureVII, FeatureVIII
from ...balderglob import RuntimeObserver


class SetupFeatureVIII(FeatureVIII):
    feature_vii = FeatureVII()

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureVIII, SetupFeatureVIII.do_something, "enter `SetupFeatureVIII.do_something`",
            category="feature")
        self.feature_vii.called_from_outer_feature()

    def called_from_outer_feature(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureVIII, SetupFeatureVIII.called_from_outer_feature,
            "enter `SetupFeatureVIII.called_from_outer_feature`", category="feature")



from __future__ import annotations
from typing import TYPE_CHECKING, Type

import itertools

from .base_connection_relation import BaseConnectionRelation, BaseConnectionRelationT

if TYPE_CHECKING:
    from ..connection import List, Union, Connection
    from .or_connection_relation import OrConnectionRelation


class AndConnectionRelation(BaseConnectionRelation):
    """
    describes an AND relation between connections
    """

    def get_tree_str(self) -> str:
        based_on_strings = [cur_elem.get_tree_str() for cur_elem in self._connections]
        return f"({' | '.join(based_on_strings)})"

    def get_possibilities_for_direct_parent_cnn(self, for_cnn_class: Type[Connection]) -> list[AndConnectionRelation]:
        """
        Helper method that returns a list of possible :class:`AndConnectionRelation` elements that hold the next parent
        in the connection-tree that can be there instead of the origin connection.
        The method returns a list, because there can exist different possibilities for this AND connection.
        """
        direct_ancestors_relations = ()
        for cur_and_elem in self.connections:
            # `cur_and_elem` needs to be a connection, because we are using simplified which has only
            # `OR[AND[Cnn, ...], Cnn, ..]`
            if cur_and_elem.__class__ in for_cnn_class.get_parents():
                # element already is a direct ancestor
                direct_ancestors_relations += (cur_and_elem,)
            else:
                all_pos_possibilities = []
                # add all possible direct parents to the possibilities list
                for cur_direct_parent in for_cnn_class.get_parents():
                    if cur_direct_parent.is_parent_of(cur_and_elem.__class__):
                        all_pos_possibilities.append(cur_direct_parent.based_on(cur_and_elem))
                direct_ancestors_relations += (all_pos_possibilities,)
        # resolve the opportunities and create multiple possible AND relations where all elements are
        # direct parents
        return [
            AndConnectionRelation(*cur_possibility).get_resolved()
            for cur_possibility in itertools.product(*direct_ancestors_relations)
        ]

    def get_simplified_relation(self) -> OrConnectionRelation:
        from ..connection import Connection  # pylint: disable=import-outside-toplevel
        from .or_connection_relation import OrConnectionRelation  # pylint: disable=import-outside-toplevel

        # create template with all elements that are definitely contained in every new AND relation (only
        # ``Connection`` objects here)
        and_template = AndConnectionRelation()
        # add all OR relations that needs to be further resolved to that list
        self_simplified_or_relations = []

        # first: go through all inner elements and convert all relations in simplified relations
        for cur_elem in self.connections:
            if isinstance(cur_elem, Connection):
                # that is fine - add it to the template
                and_template.append(cur_elem)
            elif isinstance(cur_elem, (AndConnectionRelation, OrConnectionRelation)):
                # simplify this AND/OR relation and add the items of it
                # (the simplified version is always an OR relation!)
                self_simplified_or_relations.append(cur_elem.get_simplified_relation().connections)
            else:
                raise TypeError(f'unexpected type for inner element `{cur_elem}`')
        # now our `self_simplified_or_relations` can only consist of the following nesting: `List[OR[Conn, AND[Conn]]]`
        #  we need to resolve this constellation into `OR[Conn, AND[Conn]]` now

        # now generate all possibilities out of the OR relations and add them to the AND template
        all_new_ands = []
        for cur_variation_tuple in itertools.product(*self_simplified_or_relations):
            for cur_item in cur_variation_tuple:
                if isinstance(cur_item, Connection):
                    # just add the single connection item to the AND template
                    new_full_and_relation = and_template.clone()
                    new_full_and_relation.append(cur_item)
                    all_new_ands.append(new_full_and_relation)
                elif isinstance(cur_item, AndConnectionRelation):
                    new_full_and_relation = and_template.clone()
                    new_full_and_relation.extend(cur_item.connections)
                    all_new_ands.append(new_full_and_relation)
                else:
                    #: note: inner element can not be an OR here!
                    raise TypeError(f'detect illegal type `{cur_item.__class__}` for inner element')
        if not self_simplified_or_relations:
            all_new_ands.append(and_template)
        return OrConnectionRelation(*all_new_ands)

    def is_single(self) -> bool:

        if len(self._connections) == 0:
            return True

        return min(cnn.is_single() for cnn in self._connections)

    def get_singles(self) -> List[Connection]:
        from ..connection import Connection  # pylint: disable=import-outside-toplevel

        singles_and_relations = ()
        for cur_elem in self._connections:
            # get all singles of this AND relation element
            singles_and_relations += (cur_elem.get_singles(),)
        # now get the variations and add them to our results
        return [
            Connection.based_on(AndConnectionRelation(*cur_tuple))
            for cur_tuple in itertools.product(*singles_and_relations)
        ]

    def cut_into_all_possible_subtree_branches(self) -> List[AndConnectionRelation]:
        if not self.is_single():
            raise ValueError('can not execute method, because relation is not single')

        tuple_with_all_possibilities = (
            tuple(cur_tuple_item.cut_into_all_possible_subtree_branches() for cur_tuple_item in self._connections))

        cloned_tuple_list = []
        for cur_tuple in list(itertools.product(*tuple_with_all_possibilities)):
            cloned_tuple = AndConnectionRelation(*[cur_tuple_item.clone() for cur_tuple_item in cur_tuple])
            cloned_tuple_list.append(cloned_tuple)
        return cloned_tuple_list

    def contained_in(self, other_conn: Union[Connection, BaseConnectionRelationT], ignore_metadata=False) -> bool:
        # This method checks if the AND relation is contained in the `other_conn`. To ensure that an AND relation is
        # contained in a connection tree, there has to be another AND relation into the `other_conn`, that has the same
        # length or is bigger. In addition, there has to exist an order combination where every element of the this AND
        # relation is contained in the found AND relation of the `other_cnn`. In this case it doesn't matter where the
        # AND relation is in `other_elem` (will be converted to single, and AND relation will be searched in all
        # BASED_ON elements). If the AND relation of `other_conn` has fewer items than this AND relation, it will be
        # ignored. The method only search for a valid existing item in the `other_conn` AND relation for every item of
        # this AND relation.
        from ..connection import Connection  # pylint: disable=import-outside-toplevel

        if not self.is_resolved():
            raise ValueError('can not execute method, because connection relation is not resolved')
        if not other_conn.is_resolved():
            raise ValueError('can not execute method, because other connection relation is not resolved')

        if isinstance(other_conn, BaseConnectionRelation):
            other_conn = Connection.based_on(other_conn)

        self_singles = self.get_singles()
        other_singles = other_conn.get_singles()

        for cur_self_single, cur_other_single in itertools.product(self_singles, other_singles):
            # check if we can find an AND relation in the other object -> go the single connection upwards and
            # search for a `AndConnectionRelation`

            # self is a container connection -> use raw inner AND list
            cur_self_single_and_relation = cur_self_single.based_on_elements.connections[0]

            cur_sub_other_single = cur_other_single
            while cur_sub_other_single is not None:
                if isinstance(cur_sub_other_single, AndConnectionRelation):
                    # found an AND relation -> check if length does match
                    if len(cur_sub_other_single) < len(cur_self_single_and_relation):
                        # this complete element is not possible - skip this single!
                        break
                    # length is okay, no check if every single element is contained in one of this tuple
                    for cur_inner_self_elem in cur_self_single_and_relation.connections:

                        if not cur_inner_self_elem.contained_in(cur_sub_other_single,
                                                                ignore_metadata=ignore_metadata):
                            # at least one element is not contained in other AND relation - this complete element
                            # is not possible - skip this single!
                            break
                    # all items are contained in the current other AND relation -> match
                    return True
                # go further up, if this element is no AND relation
                cur_sub_other_single = cur_sub_other_single.based_on_elements[0] \
                    if cur_sub_other_single.based_on_elements else None

        return False



from multiprocessing import Process

from _balder.balder_session import BalderSession
from _balder.testresult import ResultState


def test_1_double_illegal_vdevice_mapping(balder_working_dir):
    """
    This testcase executes the basic example and checks that the session terminates without errors and that the
    variation will not be executed.

    The test is the expanded version of the basic calculator example. Its feature `ProvidesANumberFeature` has two
    additional VDevice classes that are both mapped but on different levels (one is correctly mapped on scenario level
    and one on setup level). This is allowed, but leads to the not-executing of the variation.
    To map another vDevice this test devices an additional not connected scenario device `DoNothingDevice`.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    session.run()
    print("\n")

    assert session.executor_tree.construct_result.result == ResultState.NOT_RUN, \
            "global executor tree construct part does not set ResultState.NOT_RUN"
    assert session.executor_tree.body_result.result == ResultState.NOT_RUN, \
        "global executor tree body part does not set ResultState.NOT_RUN"
    assert session.executor_tree.teardown_result.result == ResultState.NOT_RUN, \
        "global executor tree teardown part does not set ResultState.NOT_RUN"

    assert len(session.executor_tree.get_setup_executors()) == 0, \
        "there exists some setup executor - that should not be the case"



from multiprocessing import Process

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession
from _balder.executor.unresolved_parametrized_testcase_executor import UnresolvedParametrizedTestcaseExecutor


def test_1_calculator(balder_working_dir):
    """
    This testcase executes the basic example and checks if the tree ends with the result SUCCESS
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    session.run()

    assert session.executor_tree.executor_result == ResultState.ERROR, \
        "test session does not terminates with success"

    assert session.executor_tree.construct_result.result == ResultState.SUCCESS, \
        "global executor tree construct part does not set ResultState.SUCCESS"
    assert session.executor_tree.body_result.result == ResultState.ERROR, \
        "global executor tree body part does not set ResultState.FAILURE"
    assert session.executor_tree.teardown_result.result == ResultState.NOT_RUN, \
        "global executor tree teardown part does not set ResultState.NOT_RUN"

    all_setup_executor = session.executor_tree.get_setup_executors()
    assert len(all_setup_executor) == 1, f"detect more than one setup executor"

    assert all_setup_executor[0].executor_result == ResultState.ERROR, \
        "the setup executor does not have result SUCCESS"

    assert all_setup_executor[0].construct_result.result == ResultState.SUCCESS
    assert all_setup_executor[0].body_result.result == ResultState.ERROR
    assert all_setup_executor[0].teardown_result.result == ResultState.NOT_RUN

    all_scenario_executor = all_setup_executor[0].get_scenario_executors()
    assert len(all_scenario_executor) == 1, f"detect more than one scenario executor"

    assert all_scenario_executor[0].executor_result == ResultState.ERROR, \
        "the scenario executor does not have result SUCCESS"

    assert all_scenario_executor[0].construct_result.result == ResultState.ERROR
    assert all_scenario_executor[0].body_result.result == ResultState.NOT_RUN
    assert all_scenario_executor[0].teardown_result.result == ResultState.NOT_RUN

    all_variation_executor = all_scenario_executor[0].get_variation_executors()

    assert len(all_variation_executor) == 1, f"detect more than one variation executor"

    assert all_variation_executor[0].executor_result == ResultState.NOT_RUN, \
        "the variation executor does not have result SUCCESS"

    assert all_variation_executor[0].construct_result.result == ResultState.NOT_RUN
    assert all_variation_executor[0].body_result.result == ResultState.NOT_RUN
    assert all_variation_executor[0].teardown_result.result == ResultState.NOT_RUN

    all_testcase_executor = all_variation_executor[0].get_testcase_executors()
    assert len(all_testcase_executor) == 1, f"detect some testcase executor"
    assert isinstance(all_testcase_executor[0], UnresolvedParametrizedTestcaseExecutor), \
        f"test case executor is from unexpected type {type(all_testcase_executor[0])}"

    assert all_testcase_executor[0].construct_result.result == ResultState.NOT_RUN
    assert all_testcase_executor[0].body_result.result == ResultState.NOT_RUN
    assert all_testcase_executor[0].teardown_result.result == ResultState.NOT_RUN

    exceptions = all_scenario_executor[0].get_all_recognized_exception()
    assert len(exceptions) == 1, f"detect more than one exception executor"
    assert exceptions[0].args[0] == ("feature parametrizing not possible, because "
                                     "`ScenarioAdding.NumberOneDevice.example.get_random_numbers` does not return "
                                     "Iterable")





from ...lib.features import FeatureIOverwritten
from ...balderglob import RuntimeObserver


class SetupFeatureI(FeatureIOverwritten):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureI, SetupFeatureI.do_something, "enter `SetupFeatureI.do_something`",
            category="feature")



from ...lib.features import FeatureI
from ...balderglob import RuntimeObserver


class SetupFeatureI(FeatureI):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureI, SetupFeatureI.do_something, "enter `SetupFeatureI.do_something`",
            category="feature")

    def called_from_outer_feature(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureI, SetupFeatureI.called_from_outer_feature,
            "enter `SetupFeatureI.called_from_outer_feature`", category="feature")



from typing import Union, Type
import dataclasses
from .device import Device
from .feature import Feature
from .controllers import DeviceController


class FeatureReplacementMapping:
    """
    helper object that stores mappings between scenario and setup features - is used in :class:`VariationExecutor`
    """

    @dataclasses.dataclass
    class FeatureMapping:
        """
        stores a single mapping
        """
        #: the feature attribute name in scenario device
        attr_name: str
        #: the scenario feature instance or None if the current variation does not use this setup feature in scenario
        scenario_feature: Union[Feature, None]
        #: the setup feature that is used for the scenario feature
        setup_feature: Feature

    def __init__(self):
        self._mappings: list[FeatureReplacementMapping.FeatureMapping] = []

    @property
    def mappings(self) -> list[FeatureMapping]:
        """
        returns all existing mappings
        """
        return list(self._mappings)

    @property
    def attr_names(self) -> list[str]:
        """
        returns all used attribute names
        """
        return [mapping.attr_name for mapping in self._mappings]

    @property
    def all_scenario_features(self) -> list[Feature]:
        """
        returns all scenario features
        """
        return [mapping.scenario_feature for mapping in self._mappings]

    @property
    def all_setup_features(self) -> list[Feature]:
        """
        returns all setup features
        """
        return [mapping.setup_feature for mapping in self._mappings]

    def add(self, attr_name: str, scenario_feature: Union[Feature, None], setup_feature: Feature):
        """
        adds a new mapping

        :param attr_name: the feature attribute name in scenario device
        :param scenario_feature: the scenario feature instance or None if the current variation does not use this setup
                                 feature in scenario
        :param setup_feature: the setup feature that is used for the scenario feature
        """
        if attr_name in self.attr_names:
            raise KeyError(f'entry for property name `{attr_name}` already exist - can not define it multiple times')
        self._mappings.append(FeatureReplacementMapping.FeatureMapping(attr_name, scenario_feature, setup_feature))

    def get_features_for_attr_name(self, attr_name: str) -> tuple[Feature, Feature]:
        """
        returns the scenario and its mapped setup feature for a specific attribute name used in the scenario device
        """
        for mapping in self._mappings:
            if mapping.attr_name == attr_name:
                return mapping.scenario_feature, mapping.setup_feature
        raise KeyError(f'entry for property name `{attr_name}` does not exist')

    def get_replaced_scenario_feature_for(self, setup_feature: Feature) -> Union[Feature, None]:
        """
        returns the mapped scenario feature for a given setup feature
        """
        for mapping in self._mappings:
            if mapping.setup_feature == setup_feature:
                return mapping.scenario_feature
        raise KeyError(f'cannot find setup feature for {setup_feature}')

    def add_remaining_setup_features_as_autonomous(self, cur_setup_device: Type[Device]):
        """
        Adds all features that were not added as autonomous features.
        """
        cur_setup_features = DeviceController.get_for(cur_setup_device).get_all_instantiated_feature_objects()

        # also add all setup features that are not assigned as autonomous features
        for cur_setup_feature in cur_setup_features.values():
            if cur_setup_feature not in self.all_setup_features:
                # determine free name
                idx = 0
                autonomous_name = None
                while (autonomous_name is None
                       or autonomous_name in self.attr_names):
                    autonomous_name = f"_autonomous_feat_{idx}"
                    idx += 1
                self.add(
                    attr_name=autonomous_name,
                    scenario_feature=None,
                    setup_feature=cur_setup_feature
                )



from ...lib.features import FeatureOfRelevanceLvl1, FeatureI, FeatureII, IllegalFeature
from ...balderglob import RuntimeObserver


class FeatureOfRelevanceLvl2(FeatureOfRelevanceLvl1):
    """this feature will be used nowhere"""

    class VDeviceWithI(FeatureOfRelevanceLvl1.VDeviceWithI):
        # overwrites the `FeatureI`, which is illegal because there is no relation between `IllegalFeature` and
        # `FeatureI`
        i = IllegalFeature()
        # we define this separately (to still support the requirement for a matching)
        another_one = FeatureI()

    class VDeviceWithII(FeatureOfRelevanceLvl1.VDeviceWithII):
        i = FeatureII()

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureOfRelevanceLvl2, FeatureOfRelevanceLvl2.do_something,
            "enter `FeatureOfRelevanceLvl2.do_something`", category="feature")



from ...lib.features import FeatureII
from ...balderglob import RuntimeObserver


class SetupFeatureII(FeatureII):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureII, SetupFeatureII.do_something, "enter `SetupFeatureII.do_something`",
            category="feature")
        self.feature_i.called_from_outer_feature()

    def called_from_outer_feature(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureII, SetupFeatureII.called_from_outer_feature,
            "enter `SetupFeatureII.called_from_outer_feature`", category="feature")



import balder
from ..lib.features import AddCalculateFeature, ProvidesANumberFeature, VDeviceHelperFeature1, VDeviceHelperFeature2, \
    VDeviceHelperFeature3
from ..lib.connections import SimulatedParentConnection, SimulatedChildConnection, MySimplySharedMemoryConnection


class ScenarioAdding(balder.Scenario):

    class Calculator(balder.Device):
        # we could add vDevices here later too -> f.e. `NumberOne="NumberOneDevice", NumberTwo="NumberTwoDevice"`
        adds = AddCalculateFeature()

    @balder.connect(Calculator, over_connection=balder.Connection.based_on(
        (SimulatedChildConnection.based_on(SimulatedParentConnection) & MySimplySharedMemoryConnection)))
    class NumberOneDevice(balder.Device):
        number = ProvidesANumberFeature()
        special_1 = VDeviceHelperFeature1()
        special_2 = VDeviceHelperFeature2()
        special_3 = VDeviceHelperFeature3()

    @balder.connect(Calculator, over_connection=balder.Connection)
    class NumberTwoDevice(balder.Device):
        number = ProvidesANumberFeature()

    def test_add_two_numbers(self):
        self.NumberOneDevice.number.set_number(3)
        self.NumberTwoDevice.number.set_number(4)

        self.NumberOneDevice.number.sends_the_number()
        self.NumberTwoDevice.number.sends_the_number()

        self.Calculator.adds.get_numbers()
        result = self.Calculator.adds.add_numbers()
        assert result == 7, "result is not as expected"



import balder


@balder.insert_into_tree()
class AConnection(balder.Connection):
    pass


@balder.insert_into_tree()
class BConnection(balder.Connection):
    pass


@balder.insert_into_tree()
class CConnection(balder.Connection):
    pass



from __future__ import annotations
from typing import Type, Union

import inspect

from _balder.controllers import ScenarioController
from _balder.scenario import Scenario


def covered_by(item: Union[Type[Scenario], str, callable, None]):
    """
    This decorator defines that there exists another Scenario class or test method item that has a similar
    implementation like the decorated :class:`Scenario` class or the decorated test method.

    :param item: the element which contains the whole implementation of this scenario or test method
    """

    if item is None:
        pass
    elif isinstance(item, str) and item.startswith("test_"):
        pass
    elif callable(item) and inspect.isfunction(item) and item.__name__.startswith("test_"):
        pass
    elif isinstance(item, type) and issubclass(item, Scenario):
        raise NotImplementedError('The covered-by other scenario classes is not supported yet')
    else:
        raise TypeError("the given element for `item` must be a test method of a scenario class (has to start with "
                        "`test_`)")

    class CoveredByDecorator:
        """decorator class for `@covered_by` decorator"""
        def __init__(self, func):
            self.func = func

            if inspect.isclass(func):
                # it must be a class decorator
                if not issubclass(func, Scenario):
                    raise TypeError(f"The decorator `@covered_by` may only be used for `Scenario` objects or for test "
                                    f"methods of one `Scenario` object. This is not possible for the applied class "
                                    f"`{func.__name__}`.")
                raise NotImplementedError('The covered-by decoration of other scenario classes is not supported yet')
                # scenario_controller = ScenarioController.get_for(func)
                # register for the whole class
                # scenario_controller.register_covered_by_for(meth=None, covered_by=item)
            if inspect.isfunction(func):
                # work will done in `__set_name__`
                pass
            else:
                raise TypeError(f"The use of the `@covered_by` decorator is not allowed for the `{str(func)}` element. "
                                f"You should only use this decorator for test method elements of a `Scenario` object")

        def __set_name__(self, owner, name):
            if issubclass(owner, Scenario):
                if not inspect.isfunction(self.func):
                    raise TypeError("the use of the `@covered_by` decorator is only allowed for test methods of "
                                    "`Scenario` objects")
                if not name.startswith('test_'):
                    raise TypeError(f"the use of the `@covered_by` decorator is only allowed for test methods of "
                                    f"`Scenario` objects - the method `{owner.__name__}.{name}` does not start with "
                                    f"`test_` and is not a valid test method")
                # if item is a string - resolve method
                cleared_item = item
                if isinstance(item, str):
                    cleared_item = getattr(owner, item)
                scenario_controller = ScenarioController.get_for(owner)
                scenario_controller.register_covered_by_for(meth=name, covered_by=cleared_item)
            else:
                raise TypeError(f"The use of the `@covered_by` decorator is not allowed for methods of a "
                                f"`{owner.__name__}`. You should only use this decorator for valid test methods of a "
                                f"`Scenario` object")

            setattr(owner, name, self.func)

    return CoveredByDecorator



from typing import Union


def compare_tree_output_lines(stdout_lines: list[str], expected_output: Union[list, tuple]) -> bool:

    def _check_element(remaining_stdout_lines: list[str], expected_element: Union[list, tuple, str]) -> tuple[list[str], bool]:
        if isinstance(expected_element, str):
            if expected_element == remaining_stdout_lines[0]:
                return remaining_stdout_lines[1:], True
            else:
                return remaining_stdout_lines, False
        elif isinstance(expected_element, list):
            return _check_list(remaining_stdout_lines, expected_element)
        elif isinstance(expected_element, tuple):
            return _check_tuple(remaining_stdout_lines, expected_element)
        else:
            raise TypeError(f"Expected output to be a list or tuple, got {type(expected_output)}")

    def _check_tuple(remaining_stdout_lines: list[str], expected_tuple: tuple) -> tuple[list[str], bool]:
        """tuple means, that order needs to be exact"""
        cur_remaining_stdout_lines = remaining_stdout_lines.copy()
        after_this_element_remaining_stdout_lines = remaining_stdout_lines.copy()
        for tuple_element in expected_tuple:
            after_this_element_remaining_stdout_lines, result = _check_element(cur_remaining_stdout_lines, tuple_element)
            if not result:
                return remaining_stdout_lines, False
            cur_remaining_stdout_lines = after_this_element_remaining_stdout_lines
        return after_this_element_remaining_stdout_lines, True


    def _check_list(remaining_stdout_lines: list[str], expected_list: list) -> tuple[list[str], bool]:
        """list means, that one of the element needs to pass, otherwise it should return False"""
        remaining_stdout_lines = remaining_stdout_lines.copy()
        remaining_expected_list = expected_list.copy()
        while len(remaining_expected_list) > 0:
            result_by_possibility = []

            for element in remaining_expected_list:
                result_by_possibility.append(_check_element(remaining_stdout_lines, element))

            selection = [(rem, res) for rem, res in result_by_possibility if res is True]
            if len(selection) > 1:
                raise ValueError(f'found multiple matches for list: {expected_list} within data {remaining_stdout_lines}')
            if len(selection) == 0:
                return remaining_stdout_lines, False
            remaining_stdout_lines = selection[0][0]
            remaining_expected_list.remove(
                remaining_expected_list[result_by_possibility.index((selection[0]))]
            )
        return remaining_stdout_lines, True

    full_result_remaining, full_result = _check_element(
        remaining_stdout_lines=stdout_lines.copy(),
        expected_element=expected_output.copy()
    )
    if full_result_remaining:
        return False
    return full_result



from typing import List
import balder
from ..balderglob import RuntimeObserver


class FeatureI(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureI, FeatureI.do_something, "enter `FeatureI.do_something`", category="feature")

    def get_1(self):
        return 1


class FeatureII(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureII, FeatureII.do_something, "enter `FeatureII.do_something`", category="feature")

    def get_dynamic_parameters(self, param1: int, param2: int) -> List[int]:
        raise NotImplementedError


class FeatureIII(balder.Feature):

    @property
    def params_from_property(self):
        return [1, 2, 3]

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureIII, FeatureIII.do_something, "enter `FeatureIII.do_something`", category="feature")


class FeatureIV(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureIV, FeatureIV.do_something, "enter `FeatureIV.do_something`", category="feature")

    def get_next_three_multiple_of(self, value: float) -> List[float]:
        return [value * 2, value * 3, value * 4]



from __future__ import annotations

import copy
from typing import Dict, List, Type, Union, TYPE_CHECKING

import logging
import inspect
from abc import ABC
from _balder.setup import Setup
from _balder.device import Device
from _balder.vdevice import VDevice
from _balder.scenario import Scenario
from _balder.controllers.base_device_controller import BaseDeviceController
from _balder.controllers.feature_controller import FeatureController
from _balder.exceptions import DeviceResolvingException, InnerFeatureResolvingError, \
    FeatureOverwritingError, MultiInheritanceError
if TYPE_CHECKING:
    from _balder.connection import Connection
    from _balder.controllers import ScenarioController, SetupController
    from _balder.node_gateway import NodeGateway

logger = logging.getLogger(__file__)


class DeviceController(BaseDeviceController, ABC):
    """
    This is the main device controller to manage :class:`Device` classes.
    """
    # helper property to disable manual constructor creation
    __priv_instantiate_key = object()

    #: contains all existing setup devices and its corresponding controller object
    _items: Dict[Type[Device], DeviceController] = {}

    def __init__(self, related_cls, _priv_instantiate_key):
        super().__init__()

        # this helps to make this constructor only possible inside the controller object
        if _priv_instantiate_key != DeviceController.__priv_instantiate_key:
            raise RuntimeError('it is not allowed to instantiate a controller manually -> use the static method '
                               '`DeviceController.get_for()` for it')

        if not isinstance(related_cls, type):
            raise TypeError('the attribute `related_cls` has to be a type (no object)')
        if not issubclass(related_cls, Device):
            raise TypeError(f'the attribute `related_cls` has to be a sub-type of `{Device.__name__}` but not of '
                            f'`{VDevice.__name__}`')
        if issubclass(related_cls, VDevice):
            raise TypeError(f'the attribute `related_cls` has to be a sub-type of `{Device.__name__}` but not of '
                            f'`{VDevice.__name__}`')
        if related_cls == Device:
            raise TypeError(f'the attribute `related_cls` is `{Device.__name__}` - controllers for native type are '
                            f'forbidden')
        # contains a reference to the related class this controller instance belongs to
        self._related_cls = related_cls

        # internal counter to auto provide unique node names
        self._node_cnt = 0

        #: contains all raw existing connection decorators for the related device
        self._connections: Dict[str, List[Connection]] = {}

        #: describes the absolute connections from the related device to another device
        self._absolute_connections: Dict[Type[Device], List[Connection]] = {}

        self._gateways: List[NodeGateway] = []

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    @staticmethod
    def get_for(related_cls: Type[Device]) -> DeviceController:
        """
        This class returns the current existing controller instance for the given item. If the instance does not exist
        yet, it will automatically create it and saves the instance in an internal dictionary.
        """
        if DeviceController._items.get(related_cls) is None:
            item = DeviceController(
                related_cls, _priv_instantiate_key=DeviceController.__priv_instantiate_key)
            DeviceController._items[related_cls] = item

        return DeviceController._items.get(related_cls)

    # ---------------------------------- CLASS METHODS -----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def related_cls(self) -> Type[Device]:
        """the related device type"""
        return self._related_cls

    @property
    def connections(self) -> Dict[str, List[Connection]]:
        """
        returns the defined connections for the related devices (sorted after node name) - NOT SYNCHRONIZED -
        direct decorator values
        """
        return self._connections

    @property
    def absolute_connections(self) -> Dict[Type[Device], List[Connection]]:
        """
        returns the absolute and SYNCHRONIZED connections between the related device and all other devices
        """
        return self._absolute_connections

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    def __get_outer_class_controller(self) -> Union[ScenarioController, SetupController]:
        # pylint: disable-next=import-outside-toplevel
        from _balder.controllers.normal_scenario_setup_controller import NormalScenarioSetupController

        outer_class = self.get_outer_class()
        return NormalScenarioSetupController.get_for(outer_class)

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def get_next_parent_class(self) -> Union[Type[Device], None]:
        """
        This method returns the next parent class which is a subclass of the :class:`Device` itself.

        :return: returns the next parent class or None if the next parent class is :class:`Device`
                 itself
        """
        next_base_class = None
        for cur_base in self.related_cls.__bases__:
            if issubclass(cur_base, Device):
                if next_base_class is not None:
                    raise MultiInheritanceError(
                        f"found more than one Device parent classes for `{self.related_cls.__name__}` "
                        f"- multi inheritance is not allowed for Device classes")
                next_base_class = cur_base
        if next_base_class == Device:
            return None
        return next_base_class

    def add_new_raw_connection(self, connection: Connection):
        """
        This method adds a new raw connection to the internal property `connections`.

        :param connection: the connection object (the related device has to be part of it)
        """
        if connection.from_device == self.related_cls:
            own_node = connection.from_node_name
        elif connection.to_device == self.related_cls:
            own_node = connection.to_node_name
        else:
            raise ValueError("the given connection does not have the current device as component")
        if own_node not in self._connections.keys():
            self._connections[own_node] = []

        self._connections[own_node].append(connection)

    def add_new_raw_gateway(self, gateway: NodeGateway):
        """
        This method adds a new raw gateway to the internal property `_gateways`.

        :param gateway: the gateway object (the related device has to be part of it)
        """
        if gateway.device != self.related_cls:
            raise ValueError("the given gateway does not have the current device as component")

        self._gateways.append(gateway)

    def add_new_absolute_connection(self, connection: Connection):
        """
        This method adds a new absolute connection to the internal property `absolute_connections`.

        .. note::
            This method doesn't secure any synchronizing, it only sets the connection internally.

        .. note::
            It only adds a connection, if it does not already exist in the internal list - duplicates will not be added.

        :param connection: the connection object (the related device has to be part of it)
        """
        if self.related_cls == connection.from_device:
            other_device = connection.to_device
        elif self.related_cls == connection.to_device:
            other_device = connection.from_device
        else:
            raise ValueError("the given connection does not have the current device as component")
        if other_device not in self._absolute_connections.keys():
            self._absolute_connections[other_device] = []
        if connection not in self._absolute_connections[other_device]:
            self._absolute_connections[other_device].append(connection)

    def cleanup_absolute_connections_with(self, other_device):
        """
        This method removes all connections from the related device to the given ``other_device``.
        """
        if other_device in self._absolute_connections.keys():
            del self._absolute_connections[other_device]

    def get_all_connections(self) -> Dict[str, List[Connection]]:
        """
        This method returns all available connection objects for the related device, sorted accordingly to their node
        name. The method gets all possible connection objects from the outer class (the setup or the scenario) and sorts
        the relevant ones.

        The method detects connections from other devices of the outer class that starts or ends at the related device,
        too.

        :returns: returns a mapping between the node name and a list of :class:`Connection` objects that belongs to the
                  node
        """
        outer_class_controller = self.__get_outer_class_controller()

        all_outer_class_conns = outer_class_controller.get_all_connections()
        conns_as_from_device = [
            cur_conn for cur_conn in all_outer_class_conns if cur_conn.from_device == self.related_cls]
        conns_as_to_device = [
            cur_conn for cur_conn in all_outer_class_conns if cur_conn.to_device == self.related_cls]
        result_dict = {}
        for cur_conn in conns_as_from_device:
            if cur_conn.from_node_name not in result_dict.keys():
                result_dict[cur_conn.from_node_name] = []
            result_dict[cur_conn.from_node_name].append(cur_conn)

        for cur_conn in conns_as_to_device:
            if cur_conn.to_node_name not in result_dict.keys():
                result_dict[cur_conn.to_node_name] = []
            result_dict[cur_conn.to_node_name].append(cur_conn)
        return result_dict

    def get_all_absolute_connections(self) -> Dict[str, List[Connection]]:
        """
        This method returns all available absolute connection objects for the related device, sorted accordingly to
        their node name. Absolute connection are the cleaned connections, that are reduced to work with all used
        :class:`Feature` and their :class:`VDevice` classes.

        The method also gets all possible connection objects from the outer class and sorts the relevant ones. The
        method detects connections from other devices of the outer class that starts or ends here, too.

        :returns: returns a mapping between the node name and a list of :class:`.Connection` objects that belongs to
                  the node
        """
        # we do not need to look in communication partner device here, because `_absolute_connections` is always be
        # synchronized
        outer_class_controller = self.__get_outer_class_controller()
        all_outer_class_devices = outer_class_controller.get_all_abs_inner_device_classes()

        result_dict = {}
        for cur_outer_class_device in all_outer_class_devices:
            cur_outer_class_device_controller = DeviceController.get_for(cur_outer_class_device)

            for _, cur_cnn_list in cur_outer_class_device_controller.absolute_connections.items():
                for cur_cnn in cur_cnn_list:
                    if cur_cnn.from_device == self.related_cls:
                        if cur_cnn.from_node_name not in result_dict.keys():
                            result_dict[cur_cnn.from_node_name] = []
                        if cur_cnn not in result_dict[cur_cnn.from_node_name]:
                            result_dict[cur_cnn.from_node_name].append(cur_cnn)
                    elif cur_cnn.to_device == self.related_cls:
                        if cur_cnn.to_node_name not in result_dict.keys():
                            result_dict[cur_cnn.to_node_name] = []
                        if cur_cnn not in result_dict[cur_cnn.to_node_name]:
                            result_dict[cur_cnn.to_node_name].append(cur_cnn)
        return result_dict

    # def _get_all_gateways(self) -> List[NodeGateway]:
    #     """provides a list with all gateway objects that are defined for this device"""
    #     self.__validate_gateway_node_names()
    #     return self._gateways.copy()

    # def __validate_gateway_node_names(self):
    #     """
    #     this method checks whether all `node_name` keys for the defined gateways really exist
    #     """
    #     if len(self._gateways) > 0 and self not in self._connections.keys():
    #         raise NodeNotExistsError(
    #              f"gateways are defined for non-existent nodes for the device {self.related_cls.__name__}")
    #     for cur_gateway in self._gateways:
    #         cur_gateway.validate_given_node_names()

    def get_node_types(self) -> Dict[str, List[Connection | None]]:
        """
        This method returns a dictionary with the node name as key and a connection class as value. This class
        describes the common connection sub-tree, that all incoming and outgoing connections of the related device have
        in common.

        :raises MultipleNodeBaseException: is thrown if the method finds several unrelated connections as a basis
        """

        result = {}
        all_connections = self.get_all_connections()
        for cur_node_name, node_connections in all_connections.items():
            cur_intersection = node_connections[0]
            for cur_node_connection in node_connections[1:]:
                cur_intersection = cur_node_connection.intersection_with(cur_intersection)
            result[cur_node_name] = [cur_intersection]

        return result

    def get_new_empty_auto_node(self) -> str:
        """
        This helper method returns a new empty node name. This method can be used if balder should manage node names
        automatically.
        """

        self_node_name = f"n{self._node_cnt}"
        self._node_cnt += 1
        return self_node_name

    def get_outer_class(self) -> Union[Type[Scenario], Type[Setup], None]:
        """
        This method delivers the outer class of the related device. This has to be a :class:`Setup` or a
        :class:`Scenario`.
        """
        return getattr(self.related_cls, '_outer_balder_class', None)

    def resolve_connection_device_strings(self):
        """
        This method ensures that device names, that are provided as strings within connections between the current
        device and another device (which is given as string), are resolved. Since the `@connect` marker makes it
        possible to specify the other device as a string, this method will exchange these strings with the related
        device class.

        .. note::
            This is required, because in some cases you have to provide the devices for the decorator as a string,
            because the outer class could be imported later than the execution of the decorator was done. After Balder
            has read all files, all required information are available and this method should be able to resolve the
            device-strings.
        """
        for _, node_connections in self.connections.items():
            for cur_conn in node_connections:
                # for every connection applies that the `from_device` must already be a type; also the
                # `to_device` has to be an inner class of this type

                if isinstance(cur_conn.to_device, type) and issubclass(cur_conn.to_device, Device):
                    # Skip because resolving already done
                    return

                # get outer class of `from_device`
                from_device_controller = DeviceController.get_for(cur_conn.from_device)
                parent_cls_from_device = from_device_controller.get_outer_class()

                all_inner_classes_of_outer = dict(inspect.getmembers(parent_cls_from_device, inspect.isclass))
                if cur_conn.to_device in all_inner_classes_of_outer.keys():
                    meta = cur_conn.metadata

                    to_device = all_inner_classes_of_outer[cur_conn.to_device]
                    # if there was given no unique node -> create one
                    to_device_node_name = DeviceController.get_for(to_device).get_new_empty_auto_node() \
                        if meta.to_node_name is None else meta.to_node_name

                    meta.set_to(
                        to_device=to_device,
                        to_device_node_name=to_device_node_name)

                    cur_conn.set_metadata_for_all_subitems(meta)
                else:
                    raise DeviceResolvingException(
                        f"cannot resolve the str for the given device class `{cur_conn.to_device}` for "
                        f"`@connect` decorator at device `{cur_conn.from_device.__qualname__}`")

    def validate_inner_referenced_features(self):
        """
        This method validates that every :class:`Feature` that is referenced from another :class:`Feature` of this
        device also exists in the definition list of this device.
        """
        all_instantiated_feature_objs = self.get_all_instantiated_feature_objects()
        for _, cur_feature in all_instantiated_feature_objs.items():
            cur_feature_controller = FeatureController.get_for(cur_feature.__class__)
            # now check the inner referenced features of this feature and check if that exists in the device
            for cur_ref_feature_name, cur_ref_feature in \
                    cur_feature_controller.get_inner_referenced_features().items():
                potential_candidates = []
                for _, cur_potential_candidate_feature in all_instantiated_feature_objs.items():
                    if isinstance(cur_potential_candidate_feature, cur_ref_feature.__class__):
                        # the current match is the current feature itself -> not allowed to reference itself
                        if cur_potential_candidate_feature == cur_feature:
                            raise InnerFeatureResolvingError(
                                f"can not reference the same feature from itself (done in feature "
                                f"`{cur_feature.__class__.__name__}` with `{cur_ref_feature_name}`)")
                        potential_candidates.append(cur_potential_candidate_feature)

                if len(potential_candidates) == 0:
                    raise InnerFeatureResolvingError(
                        f"can not find a matching feature in the device `{self.related_cls.__name__}` that could "
                        f"be assigned to the inner feature reference `{cur_ref_feature_name}` of the feature "
                        f"`{cur_feature.__class__.__name__}`")

                if len(potential_candidates) > 1:
                    raise InnerFeatureResolvingError(
                        f"found more than one matching feature in the device `{self.related_cls.__name__}` that "
                        f"could be assigned to the inner feature reference `{cur_ref_feature_name}` of the "
                        f"feature `{cur_feature.__class__.__name__}`")

    def validate_inheritance_of_instantiated_features(self):
        """
        This method validates instantiated features and check that they are inherited correctly. It checks that the
        feature of a child device is also a child class of the feature of the parent device (in case they use the same
        property name).
        """

        all_instantiated_feature_objs = self.get_all_instantiated_feature_objects()
        # only one match possible, because we already have checked it before
        next_base_device = self.get_next_parent_class()
        if next_base_device is not None:
            next_base_device_controller = DeviceController.get_for(next_base_device)
            # also execute this method for the base device
            next_base_device_controller.validate_inheritance_of_instantiated_features()
            all_parent_instantiated_feature_objs = next_base_device_controller.get_all_instantiated_feature_objects()
        else:
            all_parent_instantiated_feature_objs = {}

        for cur_attr_name, cur_feature in all_instantiated_feature_objs.items():
            if cur_attr_name in all_parent_instantiated_feature_objs.keys():
                # attribute name also exists before -> check if the feature is a parent of the current one
                if not isinstance(cur_feature, all_parent_instantiated_feature_objs[cur_attr_name].__class__):
                    raise FeatureOverwritingError(
                        f"the feature `{cur_feature.__class__.__name__}` with the attribute name `{cur_attr_name}` "
                        f"of the device `{self.related_cls.__name__}` you are trying to overwrite is no child class of "
                        f"the feature `{all_parent_instantiated_feature_objs[cur_attr_name].__class__.__name__}` "
                        f"that was assigned to this property before")

    def resolve_mapped_vdevice_strings(self):
        """
        This method updates the inner VDevice-Device mappings for every instantiated :class:`.Feature`, if the
        mapped device (value in constructor) was given as a string. It secures that this device has a real
        :class:`VDevice` reference for its mapped VDevice.
        """
        all_instanced_features = self.get_original_instanced_feature_objects()
        scenario_or_setup_controller = self.__get_outer_class_controller()
        if all_instanced_features is None:
            # has no features -> skip
            return
        for cur_attr_name, cur_feature in all_instanced_features.items():
            # clone feature and its active_device dict to make sure that shared instances in parent classes are handled
            # correctly
            new_feature = copy.copy(cur_feature)
            new_feature.active_vdevices = {**cur_feature.active_vdevices}
            setattr(self.related_cls, cur_attr_name, new_feature)
            if new_feature.active_vdevices != {}:
                # do something only if there exists an internal mapping
                for cur_mapped_vdevice, cur_mapped_device in new_feature.active_vdevices.items():
                    if isinstance(cur_mapped_device, str):
                        resolved_device = \
                            scenario_or_setup_controller.get_inner_device_class_by_string(cur_mapped_device)
                        if resolved_device is None:
                            raise RuntimeError(
                                f"found no possible matching name while trying to resolve "
                                f"the given vDevice string `{cur_mapped_vdevice}` in feature "
                                f"`{new_feature.__class__.__name__}`")
                        new_feature.active_vdevices[cur_mapped_vdevice] = resolved_device



from __future__ import annotations
from typing import List, Union, Dict, Type, Tuple, Iterable, TYPE_CHECKING

import copy
from _balder.connection import Connection
from _balder.node_gateway import NodeGateway
from _balder.controllers import DeviceController
from _balder.exceptions import RoutingBrokenChainError

if TYPE_CHECKING:
    from _balder.device import Device


class RoutingPath:
    """
    This object describes a possible routing path between two devices (which may be possible via several individual
    connections). If there are several possible paths between two devices, there are correspondingly many of these
    RoutingPath objects.
    """

    def __init__(self, *routing_elems: Union[Connection, NodeGateway], start_device: Type[Device],
                 start_node_name: str):
        """
        :param routing_elems: all elements that are parts of this route

        :param start_device: the device this route starts with (if the first element is a gateway this is optional)

        :param start_node_name: the node this route starts from (always required!)
        """

        self._routing_elems = []

        self._start_device = start_device
        self._start_node_name = start_node_name

        if not isinstance(routing_elems[0], Connection):
            raise TypeError("the first element is no `Connection` object - every route has to start and end with a"
                            "`Connection` object")
        if not isinstance(routing_elems[-1], Connection):
            raise TypeError("the last element is no `Connection` object - every route has to start and end with a"
                            "`Connection` object")

        first_elem = routing_elems[0]
        if isinstance(first_elem, Connection):
            if self._start_device not in (first_elem.from_device, first_elem.to_device):
                raise ValueError(f"the given `start_device={self._start_device.__name__}` does not match with one of "
                                 f"the available devices of the first routing element")
            if self._start_node_name not in (first_elem.from_node_name, first_elem.to_node_name):
                raise ValueError(f"the given `start_node_name={self._start_node_name}` does not match with one of the "
                                 f"available node names of the first routing element")
        else:
            raise TypeError(f"the first given element is not an instance of `{Connection.__name__}` or "
                            f"`{NodeGateway.__name__}` (is `{type(first_elem)}`)")

        for cur_elem in routing_elems:
            self.append_element(cur_elem)

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    @staticmethod
    def __get_abs_setup_dev_cnns_for(setup_devices: Iterable[Type[Device]]) -> List[Connection]:
        """
        Determines all absolute device connections for a given list of setup devices.
        """
        setup_devices_cnns = []
        for cur_setup_device in setup_devices:
            for cur_cnn_list in DeviceController.get_for(cur_setup_device).get_all_absolute_connections().values():
                setup_devices_cnns.extend(cur_cnn_list)
        # remove duplicates
        return list(set(setup_devices_cnns))

    @staticmethod
    def route_through(
            scenario_connection: Connection,
            device_mapping: Dict[Type[Device], Type[Device]],
            alternative_setup_device_cnns: Union[List[Connection], None] = None
    ) -> List[RoutingPath]:
        """
        This static method tries to route the given ``scenario_connection`` with the device_mapping. It returns a list
        of all matched routings between the mapped devices, where the routing is valid to support the requested
        `scenario_connection`.

        :param scenario_connection: the scenario-device connection object

        :param device_mapping: the used device mapping for the given `scenario_connection`

        :param alternative_setup_device_cnns: the alternative used connections between all relevant setup devices (if
                                              this is none, the router uses the setup-device connections from method
                                              `get_all_absolute_connections()`, otherwise it uses this dictionary here)
        """
        setup_devices_cnns = alternative_setup_device_cnns
        if alternative_setup_device_cnns is None:
            setup_devices_cnns = RoutingPath.__get_abs_setup_dev_cnns_for(device_mapping.values())

        from_setup_device = device_mapping[scenario_connection.from_device]
        to_setup_device = device_mapping[scenario_connection.to_device]

        # contains a list with all routes that start and end correctly
        all_completed_routes = []
        # contains a list with all possible routes
        all_possible_routes = []

        # add all connection objects that are mentioned in `from_setup_device`
        for cur_from_setup_node_conn in setup_devices_cnns:
            # only if there is a connection outgoing from `from_setup_device`
            if cur_from_setup_node_conn.has_connection_from_to(start_device=from_setup_device):
                all_possible_routes.append(
                    RoutingPath(
                        cur_from_setup_node_conn,
                        start_device=from_setup_device,
                        start_node_name=(cur_from_setup_node_conn.from_node_name
                                         if cur_from_setup_node_conn.from_device == from_setup_device
                                         else cur_from_setup_node_conn.to_node_name)
                    )
                )
        # now go through every possibility and add them - filter all Routes that ``has_loop() == True`` or
        # are completed
        while len(all_possible_routes) > 0:

            # remove all routings that have a loop
            all_possible_routes = [route for route in all_possible_routes.copy() if not route.has_loop()]

            # remove all not working routing because they have the wrong connection type, by checking that one part
            # connection matches the requirements of the given `scenario_connection`
            all_possible_routes = [
                r for r in all_possible_routes
                if scenario_connection.contained_in(r.get_virtual_connection(), ignore_metadata=True)
            ]

            # move all completely routed connections to `all_completed_routes`
            for cur_routing in all_possible_routes.copy():
                if cur_routing.end_device == to_setup_device:
                    # note that all routes already have a virtual connection that matches the requirements of given
                    # `scenario_connection`
                    all_completed_routes.append(cur_routing)
                    all_possible_routes.remove(cur_routing)

            new_possible_routes = []
            for cur_routing in all_possible_routes.copy():
                # add all existing connections
                all_next_conns = [cur_cnn for cur_cnn in setup_devices_cnns
                                  if ((cur_cnn.from_device == cur_routing.end_device and
                                       cur_cnn.from_node_name == cur_routing.end_node_name)
                                      or (cur_cnn.to_device == cur_routing.end_device and
                                          cur_cnn.to_node_name == cur_routing.end_node_name))]
                for cur_next_conn in all_next_conns:
                    if cur_next_conn == cur_routing.elements[-1]:
                        # is the same connection as the last of routing -> SKIP
                        continue

                    if cur_next_conn.has_connection_from_to(start_device=cur_routing.end_device):
                        # the connection allows the direction the routing needs - only then add it
                        copied_routing = cur_routing.copy()
                        copied_routing.append_element(cur_next_conn)
                        new_possible_routes.append(copied_routing)
                # add all possible gateways
                # for cur_next_gateway in cur_routing.end_device._get_all_gateways():
                #     if cur_next_gateway == cur_routing.elements[-1]:
                #         # is the same gateway as the last one of the routing -> SKIP
                #         continue
                #     elif cur_next_gateway.has_connection_from_to(start_node=cur_routing.end_node_name):
                #         # the gateway allows the direction the routing needs - only then add it
                #         copied_routing = cur_routing.copy()
                #         copied_routing.append_element(cur_next_gateway)
                #         all_possible_routes.append(copied_routing)
            all_possible_routes = new_possible_routes

        return all_completed_routes

    # ---------------------------------- CLASS METHODS ----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def elements(self) -> List[Union[Connection, NodeGateway]]:
        """returns all elements that belongs to this routing path"""
        return self._routing_elems

    @property
    def start_device(self) -> Type[Device]:
        """returns the device the route starts from"""
        return self._start_device

    @property
    def start_node_name(self) -> str:
        """returns the node of the `start_device` this route starts from"""
        return self._start_node_name

    @property
    def end_device(self) -> Type[Device]:
        """returns the device the route ends"""
        return self._get_end_device_and_node()[0]

    @property
    def end_node_name(self) -> str:
        """returns the node of the `end_device` this route ends"""
        return self._get_end_device_and_node()[1]

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    def _get_end_device_and_node(self, till_idx: int = None) -> Tuple[Type[Device], str]:
        """
        helper method that determines the end_device and end_node_name

        :param till_idx: the index of the latest element that should be considered

        :return: a tuple with the latest device and node
        """
        cur_device = self._start_device
        cur_node_name = self._start_node_name

        elements = self._routing_elems
        if till_idx is not None:
            elements = self._routing_elems[:till_idx+1]

        for cur_route_elem in elements:
            if isinstance(cur_route_elem, NodeGateway):
                if cur_node_name == cur_route_elem.from_node_name:
                    cur_node_name = cur_route_elem.to_node_name
                elif cur_node_name == cur_route_elem.to_node_name:
                    cur_node_name = cur_route_elem.from_node_name
                else:
                    raise RoutingBrokenChainError(
                        f"can not chain the routing element `{cur_route_elem.__name__}` with the device "
                        f"`{cur_device.__name__}` and node `{cur_node_name}` of element before")
            else:
                if cur_node_name == cur_route_elem.from_node_name and cur_device == cur_route_elem.from_device:
                    cur_node_name = cur_route_elem.to_node_name
                    cur_device = cur_route_elem.to_device
                elif cur_node_name == cur_route_elem.to_node_name and cur_device == cur_route_elem.to_device:
                    cur_node_name = cur_route_elem.from_node_name
                    cur_device = cur_route_elem.from_device
                else:
                    raise RoutingBrokenChainError(
                        f"can not chain the routing element `{cur_route_elem.__name__}` with the device "
                        f"`{cur_device.__name__}` and node `{cur_node_name}` of element before")
        return cur_device, cur_node_name

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def has_loop(self) -> bool:
        """
        This method returns True if it detects an internal loop. An internal loop is given, if one :class:`Device`/node
        pair is mentioned twice (or more) in internal `routing_elements`.
        """
        all_contact_points = [(self.start_device, self.start_node_name)]

        for idx in range(len(self._routing_elems)):
            next_device, next_node = self._get_end_device_and_node(idx)
            # now check if one point is mentioned twice
            if (next_device, next_node) in all_contact_points:
                return True
            all_contact_points.append((next_device, next_node))

        return False

    def is_bidirectional(self):
        """
        This method checks whether the route is completely bidirectional. If only one connection is not bidirectional,
        it will return False.
        """
        for cur_elem in self._routing_elems:
            if isinstance(cur_elem, Connection):
                if not cur_elem.is_bidirectional():
                    return False
            else:
                if not cur_elem.is_bidirectional():
                    return False
        return True

    def copy(self) -> RoutingPath:
        """
        This method creates a copy of this routing.
        """
        copied_elem = copy.copy(self)
        # also copy list reference
        copied_elem._routing_elems = self._routing_elems.copy()  # pylint: disable=protected-access
        return copied_elem

    def append_element(self, elem: Union[Connection, NodeGateway]) -> None:
        """
        This method appends an element to the route.
        """
        elem_before = None if len(self._routing_elems) == 0 else self._routing_elems[-1]

        if not isinstance(elem, Connection) and not isinstance(elem, NodeGateway):
            raise TypeError(f"the given attribute passed at position {len(self._routing_elems)} is not of the type "
                            f"`Connection` or `NodeGateway`")

        if elem_before is not None:
            # check if device and nodes of the two elements are the same -> have to be a chain
            if isinstance(elem, Connection):
                if (self.end_device, self.end_node_name) not in \
                        [(elem.from_device, elem.from_node_name), (elem.to_device, elem.to_node_name)]:
                    raise RoutingBrokenChainError(
                        f"can not append connection, because neither the from-device/node "
                        f"(device: `{elem.from_device.__name__}` | node: `{elem.from_node_name}`) nor the "
                        f"to-device/node (device: `{elem.to_device.__name__}` | node: `{elem.to_node_name}`) of the "
                        f"connection match with the latest end-device/node (device: `{self.end_device.__name__}` | "
                        f"node: `{self.end_node_name}`) of this route")
            else:
                # is a gateway class
                if self.end_device != elem.device:
                    raise RoutingBrokenChainError(
                        f"can not append gateway, because the device of the gateway (`{elem.device}`) doesn't "
                        f"match with the latest end-device (`{self.end_device}`) of this route")
                if self.end_node_name not in [elem.from_node_name, elem.to_node_name]:
                    raise RoutingBrokenChainError(
                        f"can not append gateway, because neither the from-node "
                        f"(`{elem.from_node_name}`) nor the to-node (`{elem.to_node_name}`) of the gateway match with "
                        f"the latest end-node (`{self.end_node_name}`) of this route")

        self._routing_elems.append(elem)

    def get_virtual_connection(self) -> Connection:
        """
        This method returns a virtual connection object that describes the connection type this routing
        supports for all of its elements.
        """
        virtual_connection = self.elements[0].clone()
        virtual_connection.set_metadata_for_all_subitems(None)
        for cur_element in self.elements[1:]:
            if isinstance(cur_element, Connection):
                cur_element_clone = cur_element.clone()
                cur_element_clone.set_metadata_for_all_subitems(None)
                virtual_connection = cur_element_clone.intersection_with(virtual_connection)
            else:
                # is a gateway class
                # todo
                pass
        # set metadata based on this routing
        virtual_connection.metadata.set_from(from_device=self.start_device, from_device_node_name=self.start_node_name)
        virtual_connection.metadata.set_to(to_device=self.end_device, to_device_node_name=self.end_node_name)

        return virtual_connection



from __future__ import annotations

from _balder.connection import Connection
from _balder.decorator_insert_into_tree import insert_into_tree

from _balder.objects.connections import osi_1_physical


@insert_into_tree(parents=[osi_1_physical.OpticalFiberConnection, osi_1_physical.DslConnection,
                           osi_1_physical.IsdnConnection])
class EthernetConnection(Connection):
    """
    Balder Ethernet connection (OSI LAYER 2)
    """


@insert_into_tree(parents=[osi_1_physical.WifiConnection])
class WirelessLanConnection(Connection):
    """
    Balder Wireless LAN connection (OSI LAYER 2)
    """


@insert_into_tree(parents=[])
class LLDPConnection(Connection):
    """
    Balder LLDP connection (OSI LAYER 2)
    """


@insert_into_tree(parents=[])
class ProfibusConnection(Connection):
    """
    Balder Profibus connection (OSI LAYER 2)
    """



from _balder.routing_path import RoutingPath
from balder.exceptions import RoutingBrokenChainError
import balder


class Device1(balder.Device):
    pass


class Device2(balder.Device):
    pass


class Device3(balder.Device):
    pass


class Device4(balder.Device):
    pass


def test_simple_valid_route_over_3_devs():
    """
    This test validates the :class:`RoutingPath` for a valid connection from `Device1` -> `Device2` -> `Device3`.
    """

    cnn10_20 = balder.Connection(
        from_device=Device1, from_device_node_name='0', to_device=Device2, to_device_node_name='0')
    cnn20_30 = balder.Connection(
        from_device=Device2, from_device_node_name='0', to_device=Device3, to_device_node_name='0')
    cnn10_30 = balder.Connection(
        from_device=Device1, from_device_node_name='0', to_device=Device3, to_device_node_name='0')

    path = RoutingPath(cnn10_20, start_device=Device1, start_node_name='0')
    path.append_element(cnn20_30)

    assert path.get_virtual_connection() == cnn10_30


def test_simple_valid_route_over_3_devs_with_inverse_def():
    """
    This test validates the :class:`RoutingPath` for a valid connection from `Device1` -> `Device2` -> `Device3`.
    The second connection is defined inverse, which means, that the from-device is `Device3` instead of `Device1`. As
    all connection are bidirectional, this should work without any problems.
    """

    cnn10_20 = balder.Connection(
        from_device=Device1, from_device_node_name='0', to_device=Device2, to_device_node_name='0')
    cnn30_20 = balder.Connection(
        from_device=Device3, from_device_node_name='0', to_device=Device2, to_device_node_name='0')
    cnn10_30 = balder.Connection(
        from_device=Device1, from_device_node_name='0', to_device=Device3, to_device_node_name='0')

    path = RoutingPath(cnn10_20, start_device=Device1, start_node_name='0')
    path.append_element(cnn30_20)

    assert path.get_virtual_connection() == cnn10_30


def test_invalid_route_wrong_node():
    """
    This test checks that the method :meth:`RoutingPath.append_element` throws a :class:`RoutingBrokenChainError`
    if the device-node-name is not the same as the end-node-name of the last element before the new connection is added.
    """

    cnn10_20 = balder.Connection(
        from_device=Device1, from_device_node_name='0', to_device=Device2, to_device_node_name='0')
    cnn21_30 = balder.Connection(
        from_device=Device2, from_device_node_name='1', to_device=Device3, to_device_node_name='0')

    path = RoutingPath(cnn10_20, start_device=Device1, start_node_name='0')
    try:
        path.append_element(cnn21_30)
    except RoutingBrokenChainError as exc:
        assert exc.args[0] == 'can not append connection, because neither the from-device/node ' \
                              '(device: `Device2` | node: `1`) nor the to-device/node ' \
                              '(device: `Device3` | node: `0`) of the connection match with the latest ' \
                              'end-device/node (device: `Device2` | node: `0`) of this route'


def test_invalid_route_wrong_node_with_inverse_def():
    """
    This test checks that the method :meth:`RoutingPath.append_element` throws a :class:`RoutingBrokenChainError`
    if the device-node-name is not the same as the end-node-name of the last element before the new connection is added.

    The second connection is defined inverse, which means, that the from-device is `Device3` instead of `Device1`. As
    all connection are bidirectional, this should also throw the wrong node error.
    """

    cnn10_20 = balder.Connection(
        from_device=Device1, from_device_node_name='0', to_device=Device2, to_device_node_name='0')
    cnn30_21 = balder.Connection(
        from_device=Device3, from_device_node_name='0', to_device=Device2, to_device_node_name='1')

    path = RoutingPath(cnn10_20, start_device=Device1, start_node_name='0')
    try:
        path.append_element(cnn30_21)
    except RoutingBrokenChainError as exc:
        assert exc.args[0] == 'can not append connection, because neither the from-device/node ' \
                              '(device: `Device3` | node: `0`) nor the to-device/node ' \
                              '(device: `Device2` | node: `1`) of the connection match with the latest ' \
                              'end-device/node (device: `Device2` | node: `0`) of this route'


def test_invalid_route_no_dev():
    """
    This test checks, that the method :meth:`RoutingPath.append_element` throws a :class:`RoutingBrokenChainError` in
    case a connection will be added that does not have the current end-device.
    """

    cnn10_20 = balder.Connection(
        from_device=Device1, from_device_node_name='0', to_device=Device2, to_device_node_name='0')
    cnn30_40 = balder.Connection(
        from_device=Device3, from_device_node_name='0', to_device=Device4, to_device_node_name='0')

    path = RoutingPath(cnn10_20, start_device=Device1, start_node_name='0')
    try:
        path.append_element(cnn30_40)
    except RoutingBrokenChainError as exc:
        assert exc.args[0] == 'can not append connection, because neither the from-device/node ' \
                              '(device: `Device3` | node: `0`) nor the to-device/node ' \
                              '(device: `Device4` | node: `0`) of the connection match with the latest ' \
                              'end-device/node (device: `Device2` | node: `0`) of this route'


def test_invalid_route_diff_dev_and_diff_node():
    """
    This test checks, that the method :meth:`RoutingPath.append_element` throws a :class:`RoutingBrokenChainError` in
    case a connection will be added that does not have the current end-device and current end-device-node-name.
    """

    cnn10_20 = balder.Connection(
        from_device=Device1, from_device_node_name='0', to_device=Device2, to_device_node_name='0')
    cnn31_41 = balder.Connection(
        from_device=Device3, from_device_node_name='1', to_device=Device4, to_device_node_name='1')

    path = RoutingPath(cnn10_20, start_device=Device1, start_node_name='0')
    try:
        path.append_element(cnn31_41)
    except RoutingBrokenChainError as exc:
        assert exc.args[0] == 'can not append connection, because neither the from-device/node ' \
                              '(device: `Device3` | node: `1`) nor the to-device/node ' \
                              '(device: `Device4` | node: `1`) of the connection match with the latest ' \
                              'end-device/node (device: `Device2` | node: `0`) of this route'
